# Copyright (C) 2014 MediaMath, Inc. <http://www.mediamath.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re
import sqlite3
import time

from twisted.internet.protocol import Factory
from twisted.protocols.basic import LineReceiver

import util

class SqlReceiver(LineReceiver):
    """
    A simple line based sql receiver.
    """

    def __init__(self, factory, connection_id):
        self.sql_statement = ''
        self.nr_queries = 0
        self.factory = factory
        self.connection_id = connection_id


    def lineReceived(self, line):
        """
        When a line comes in from the client appened it to the working sql statement.
        If its a complete statement (semicolon terminated), execute it.
        """

        # Exit on lines with just ctrl-d and/or ctrl-c

        m = re.search(r"^\s*[" + chr(4) + chr(3) + "]+\s*$", line, flags=re.IGNORECASE)

        if m != None:
            self.sendLine("Bye!")
            return 1

        # Add this line to the multi-line sql statement.

        self.sql_statement += line + "\n"

        # Do we have a complete sql statement?

        if sqlite3.complete_statement(self.sql_statement):
            
            query_id = self.factory.data_manager.get_query_id()

            query_start = time.time()

            logging.info("SqlReceiver: (%d:%d) SQL received: %s", self.connection_id, query_id, self.sql_statement.rstrip())

            
            # Enqueue
            d = self.factory.data_manager.async_validate_and_route_query(self.sql_statement, query_id)

            d.addCallback(self.sql_complete_callback, query_id, query_start)


    def sql_complete_callback(self, result, query_id, query_start):
        """
        Called when a sql statement completes.
        """

        retval = result["retval"]
        data          = result["data"]          if "data"          in result else None
        error_message = result["error_message"] if "error_message" in result else ''
        max_widths    = result["max_widths"]    if "max_widths"    in result else None

        if retval == 0 and data != None:
            util.pretty_print_table(self, data, max_widths=max_widths)

        if retval != 0:
            logging.info("SqlReceiver: (%d:%d) SQL error: %s", self.connection_id, query_id, error_message)
            self.sendLine("SQL error: %s" % error_message)
        else:
            logging.info("SqlReceiver: (%d:%d) SQL completed (%.02f seconds)", self.connection_id, query_id, time.time() - query_start)

        self.nr_queries += 1
        self.factory.nr_queries += 1
        self.sql_statement = ''



    def connectionMade(self):
        logging.info("SqlReceiver: (%d) Connection received!", self.connection_id)

    def connectionLost(self, reason):
        if self.sql_statement != '':
            logging.info("SqlReceiver: (%d) Connection terminated! %d queries SQL: \n%s", self.connection_id, self.nr_queries, self.sql_statement)
        else:
            logging.info("SqlReceiver: (%d) Connection terminated! %d queries", self.connection_id, self.nr_queries)



class SqlReceiverFactory(Factory):

    def __init__(self, data_manager):
        self.connection_id = 0
        self.nr_statements = 0
        self.data_manager = data_manager
        self.nr_queries = 0

    def set_backend_connection(self, backend_connection):
        self.backend_connection = backend_connection

    def buildProtocol(self, addr):
        self.connection_id += 1
        return SqlReceiver(self, self.connection_id)
        #return self.data_manager.get_sql_receiver(self, self.connection_id)




#!/usr/bin/python

# Copyright (C) 2014 MediaMath, Inc. <http://www.mediamath.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import logging
from pprint import pprint
from optparse import OptionParser
import re
import random
import time

from twisted.internet import reactor
from twisted.internet import task


for path in [
    os.path.join('opt', 'qasino', 'lib'),
    os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'lib'))
]:
    if os.path.exists(os.path.join(path, '__init__.py')):
        sys.path.insert(0, path)
        break

from txzmq import ZmqFactory

from csv_table_reader import CsvTableReader
import constants
from util import Identity
import qasino_table


def main():

    global options

    logging.basicConfig(format="%(asctime)s %(message)s", datefmt="%Y-%m-%d %H:%M:%S",
                        level=logging.INFO)

    parser = OptionParser()

    parser.add_option("-I", "--identity", dest="identity",
                      help="Use IDENTITY as identity", metavar="IDENTITY")

    parser.add_option("-H", "--hostname", dest="hostname", default='localhost',
                      help="Send table to HOSTNAME qasino server", metavar="HOSTNAME")

    parser.add_option("-p", "--port", dest="port", default=constants.ZMQ_RPC_PORT, type=int,
                      help="Use PORT for qasino server", metavar="PORT")

    parser.add_option("-u", "--username", dest="username", 
                      help="HTTPS auth username")

    parser.add_option("-w", "--password", dest="password", 
                      help="HTTPS auth password")

    parser.add_option("-P", "--pubsub-port", dest="pubsub_port", default=constants.ZMQ_PUBSUB_PORT, type=int,
                      help="Use PORT for qasino pubsub connection", metavar="PORT")

    parser.add_option("-i", "--index", dest="indexes",
                      action="append",
                      help="Path to a index file to process" )

    parser.add_option("-f", "--index-list", dest="index_list",
                      help="Path to a file with a list of index files to process in it" )

    parser.add_option("-T", "--table", dest="tables",
                      action="append",
                      help="Tables to limit publishing to" )

    parser.add_option("-t", "--table-list", dest="table_list",
                      help="Path to a file with a list of tables to limit publishing to" )

    parser.add_option("-d", "--send-delay-max", dest="send_delay_max", default=15,
                      help="Max delay to add when its time to send tables." )

    parser.add_option("-x", "--interval", dest="interval", default=None, type=int,
                      help="Interval to send updates (This will turn off subscribing)." )

    parser.add_option("-s", "--use-https", dest="use_https", default=False, action="store_true",
                      help="Use HTTP over SSL/TLS protocol to publish table.")

    parser.add_option("-k", "--skip-ssl-verify", dest="skip_ssl_verify", default=False, action="store_true",
                      help="Don't verify SSL certificates.")

    parser.add_option("-g", "--gen-signal-timeout", dest="gen_signal_timeout", default=120, type=int,
                      help="Timeout after which we restart the generation signal subscription.")

    (options, args) = parser.parse_args()

    logging.info("Qasino csv publisher starting")

    if options.identity != None:
        Identity.set_identity(options.identity)

    logging.info("Identity is %s", Identity.get_identity())

    if options.hostname == None:
        logging.info("Please specify a hostname to connect to.")
        exit(1)

    zmq_factory = ZmqFactory()

    # Create a request object (either ZMQ or HTTPS).

    if options.use_https:
        import http_requestor

        # Change the default port if we're https
        if options.port == constants.ZMQ_RPC_PORT:
            options.port = constants.HTTPS_PORT

        logging.info("Connecting to {}:{} with HTTPS to send tables.".format(options.hostname, options.port))

        # Disable extraneous logging in requests.
        requests_log = logging.getLogger("requests")
        requests_log.setLevel(logging.WARNING)

        requestor = http_requestor.HttpRequestor(options.hostname, options.port, 
                                                 username = options.username,
                                                 password = options.password, 
                                                 skip_ssl_verify = options.skip_ssl_verify )
    else:  # Use zmq requestor
        import zmq_requestor

        logging.info("Connecting to {}:{} with ZeroMQ to send tables.".format(options.hostname, options.port))

        requestor = zmq_requestor.ZmqRequestor(options.hostname, options.port, zmq_factory)

    # Determine the update trigger (interval or signal).

    if options.interval is None or options.interval < 10:

        logging.info("Connecting to {}:{} on pubsub ZeroMQ channel to listen for generation signals.".format(options.hostname, options.pubsub_port))

        # Create a zeromq pub sub subscriber.

        import zmq_subscriber

        subscriber = zmq_subscriber.ZmqSubscriber(options.hostname, options.pubsub_port, zmq_factory)

        # Read and send the table when a generation signal comes in.

        global last_gen_signal_time

        last_gen_signal_time = time.time()

        subscriber.subscribe_generation_signal(initiate_read_and_send_tables, requestor, options)

        # Set a timeout so we can restart the subscribe if we haven't heard from the server in a while.

        reactor.callLater(5, check_for_gen_signal_timeout, options, subscriber, requestor, zmq_factory)

    else:
        # Read and send the table at a fixed interval.

        logging.info("Sending data on fixed interval ({} seconds).".format(options.interval))

        request_metadata_task = task.LoopingCall(read_and_send_tables, requestor, options)
        request_metadata_task.start(int(options.interval))


    # Read and send immediately, uncomment.

#    read_and_send_tables(requestor, options)

    # Run the event loop

    reactor.run()

    logging.info("Qasino csv publisher exiting")


def check_for_gen_signal_timeout(options, prev_subscriber, requestor, zmq_factory):

    global last_gen_signal_time

    if time.time() - last_gen_signal_time > options.gen_signal_timeout:
        logging.warning("Exceeded the generation signal timeout (%ds).  Restarting ZeroMQ subscriber.", options.gen_signal_timeout)
        
        import zmq_subscriber

        prev_subscriber.shutdown()

        subscriber = zmq_subscriber.ZmqSubscriber(options.hostname, options.pubsub_port, zmq_factory)

        # Read and send the table when a generation signal comes in.

        subscriber.subscribe_generation_signal(initiate_read_and_send_tables, requestor, options)

        # Reset the timer and last signal time so we don't re-subscribe right away.
        last_gen_signal_time = time.time()
        reactor.callLater(5, check_for_gen_signal_timeout, options, subscriber, requestor, zmq_factory)
    else:

        # Reset the timer to check again in a bit.
        reactor.callLater(5, check_for_gen_signal_timeout, options, prev_subscriber, requestor, zmq_factory)
        


def get_csv_files_from_index(index_file):
    """
    Read the csv filenames from an index file.
    Returns a list of pairs of filename and tablename.
    """

    results = []

    try:
        fh = open(index_file, 'r')
    except Exception as e:
        logging.info("Failed to open index file '%s': %s", index_file, e)
        return

    for lineno, line in enumerate(fh):

        # Skip the first line, its a useless version number.

        if lineno == 0:
            continue

        # Remove newlines at the end.
        line = line.rstrip('\r\n')

        # If the line contains a comma (,) then it has filename and tablename.

        fields = line.split(',')

        if len(fields) > 1:
            filename = fields[0]
            tablename = fields[1]

        else:
            tablename = fields[0]
            filename = fields[0] + ".csv"
            
        results.append( [ filename, tablename ] )

    fh.close()

    return results


def get_index_list_file_indexes(index_list_file):
    """
    Reads an "index list file" to get a list if indexes to process.
    Returns a list of index file paths.
    """

    result_indexes = []

    # Read the file - each line is the full path to an index file.

    try:
        fh = open(index_list_file, "r")
    except Exception as e:
        logging.info("Failed to open index list file '%s': %s", index_list_file, e)
        return

    for line in fh:

        # Skip comments
        m = re.match(r'\s*#', line)
        if m: continue
        
        # Skip blank lines
        m = re.match(r'\s*$', line)
        if m: continue

        # Remove newlines at the end.
        line = line.rstrip("\r\n")

        result_indexes.append(line)

    fh.close()

    return result_indexes


def get_table_list_file_tables(table_list_file):
    """
    Reads a "table list file" to get a list if tables to limit to.
    Returns a list of table names.
    """

    result_tables = []

    # Read the file - each line is the tablename

    try:
        fh = open(table_list_file, "r")
    except Exception as e:
        logging.info("Failed to open table list file '%s': %s", table_list_file, e)
        return

    for line in fh:

        # Skip comments
        m = re.match(r'\s*#', line)
        if m: continue
        
        # Skip blank lines
        m = re.match(r'\s*$', line)
        if m: continue

        # Remove newlines at the end.
        line = line.rstrip("\r\n")

        result_tables.append(line)

    fh.close()

    return result_tables


def initiate_read_and_send_tables(requestor, options):
    """ 
    Calls read_and_send_tables after a random delay to reduce "storming" the server.
    """
    global last_gen_signal_time

    last_gen_signal_time = time.time()

    delay = random.randint(0, int(options.send_delay_max))
    logging.info("Waiting %d seconds to send data.", delay)
    reactor.callLater(delay, read_and_send_tables, requestor, options)

def get_mtime(filepath):
    try:
        return os.path.getmtime(filepath)
    except:
        return 0


def read_and_send_tables(requestor, options):
    """
    Given the specified indexes, read the csv files and publish them
    to a qasino server.
    """

    nr_tables = 0
    nr_errors = 0
    table_info = {}

    # Make a table whitelist lookup dict.
    table_whitelist = {}
    use_table_whitelist = False

    # Was there one more tables given on the command line?

    if options.tables and len(options.tables) > 0:
        use_table_whitelist = True
        for x in options.tables:
            table_whitelist[x] = 1

    # Was there a table list file given on the command line?

    if options.table_list != None and len(options.table_list) > 0:

        tables_from_file = get_table_list_file_tables(options.table_list)

        if tables_from_file and len(tables_from_file) > 0:
            use_table_whitelist = True
            for x in tables_from_file:
                table_whitelist[x] = 1


    # Create a csv table reader object.

    csv_table_reader = CsvTableReader()

    # This will be the list of indexes to process.
    indexes = []

    # Was there one or more index files given on the command line?

    if options.indexes: 
        indexes = indexes + options.indexes

    # Was there an index list file given on the command line?

    if options.index_list:

        indexes_from_file = get_index_list_file_indexes(options.index_list)

        if indexes_from_file:
            indexes = indexes + indexes_from_file


    # Now process all the indexes.

    for index_file in indexes:

        index_dir = os.path.dirname(index_file)

        csv_files = get_csv_files_from_index(index_file)

        if csv_files is None or len(csv_files) <= 0:

            logging.info("Warning: no csv files found in index '%s'", index_file)
            continue

        for csv_file_item in csv_files:

            (filename, tablename) = csv_file_item

            # Is this a tablename we can process?

            if use_table_whitelist:
                if tablename not in table_whitelist:
                    continue

            filepath = '/'.join( [index_dir, filename] )

            logging.info("Reading file '%s'.", filepath)

            # The csv files we'll be reading in will have this format:
            # ------------------------
            # version/options
            # tablename
            # column names (csv)
            # column types (csv)
            # column descriptions (csv)
            # data (csv)
            # ...
            # ------------------------

            table_info[tablename] = {}
            table_info[tablename]["filepath"] = filepath
            table_info[tablename]["nr_rows"] = -1
            table_info[tablename]["nr_errors"] = 0
            table_info[tablename]["error_msg"] = ''
            table_info[tablename]["read_epoch"] = time.time()
            table_info[tablename]["mtime"] = get_mtime(filepath)

            try:
                filehandle = open(filepath, 'r')
            except Exception as e:
                nr_errors += 1
                table_info[tablename]["nr_errors"] = 1
                table_info[tablename]["error_msg"] = e.str()
                logging.info("Failure reading csv file '%s': %s", filepath, error)
                continue

            # Ignore the 2nd and 5th lines.  Names in 3rd, types in 4th.  The "version" in the first line is now the options list.
            (table, error) = csv_table_reader.read_table(filehandle, tablename,
                                                         skip_linenos={1, 4},
                                                         options_lineno=0,
                                                         types_lineno=3,
                                                         colnames_lineno=2)

            filehandle.close()

            table_info[tablename]["read_time_s"] = time.time() - table_info[tablename]["read_epoch"]

            if table == None:
                nr_errors += 1
                table_info[tablename]["nr_errors"] = 1
                table_info[tablename]["error_msg"] = error
                logging.info("Failure reading csv file '%s': %s", filepath, error)
                continue

            nr_tables += 1

            table_info[tablename]["nr_rows"] = table.get_nr_rows()

            properties = []
            if table.get_property('static'): properties.append(' static')
            if table.get_property('update'): properties.append(' update')
            if table.get_property('persist'): properties.append(' persist')

            logging.info("Sending{} table '{}' to '{}:{}' ({} rows).".format( ''.join(properties), tablename, options.hostname, options.port, table_info[tablename]["nr_rows"] ) )

            error = requestor.send_table(table)

            if error is not None:
                logging.info("Error sending table '{}': {}".format(tablename, error))
        
        # END for each csv file

    # END for each index

    # Publish an info table

    publish_info_table(requestor, nr_tables, nr_errors)

    # Publish a table list table.

    publish_tables_table(requestor, table_info)


def publish_info_table(requestor, nr_tables, nr_errors):

    tablename = "qasino_csvpublisher_info"

    table = qasino_table.QasinoTable(tablename)
    table.add_column("identity", "varchar")
    table.add_column("update_epoch", "int")
    table.add_column("nr_tables", "int")
    table.add_column("nr_errors", "int")
    table.add_row( [ Identity.get_identity(), time.time(), nr_tables, nr_errors ] )

    logging.info("Sending table '%s' to '%s:%d' (1 rows).", tablename, options.hostname, options.port)

    requestor.send_table(table)

def publish_tables_table(requestor, table_info):

    this_tablename = "qasino_csvpublisher_tables"

    table = qasino_table.QasinoTable(this_tablename)
    table.add_column("identity", "varchar")
    table.add_column("tablename", "varchar")
    table.add_column("read_epoch", "real")
    table.add_column("read_time_s", "int")
    table.add_column("mtime", "int")
    table.add_column("nr_errors", "int")
    table.add_column("error_msg", "int")
    table.add_column("nr_rows", "int")
    table.add_column("filepath", "varchar")

    for tablename, table_stats in table_info.iteritems():
        table.add_row( [ Identity.get_identity(), 
                         tablename,
                         table_stats["read_epoch"], 
                         table_stats["read_time_s"], 
                         table_stats["mtime"], 
                         table_stats["nr_errors"],
                         table_stats["error_msg"],
                         table_stats["nr_rows"],
                         table_stats["filepath"] ] )
        
    logging.info("Sending table '%s' to '%s:%d' (%d rows).", this_tablename, options.hostname, options.port, table.get_nr_rows())

    requestor.send_table(table)



if __name__ == '__main__':

    exit( main() )



# Copyright (C) 2014 MediaMath, Inc. <http://www.mediamath.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import csv
import re

import logging

import qasino_table

class CsvTableReader(object):

    csv_to_qasino_type_map = { 'str' : 'TEXT',
                               'string' : 'TEXT',
                               'ip' : 'TEXT',
                               'varchar' : 'TEXT',
                               'float' : 'REAL',
                               'int' : 'INTEGER',
                               'integer' : 'INTEGER',
                               'll' : 'INTEGER',
                               'time' : 'INTEGER'
                               }

    def istrue(self, str):
        if str == None:
            return False

        m = re.search(r'^(false|no|0)$', str, flags=re.IGNORECASE)
        if m != None:
            return False

        return True


    def read_table(self, filehandle, tablename, 
                   colnames_lineno=1,
                   types_lineno=2,
                   options_lineno=-1,
                   tablename_lineno=-1,
                   skip_linenos=set()):

        table = qasino_table.QasinoTable(tablename)

        column_names = None
        column_types = None
    
        try:
            for lineno, line in enumerate(filehandle):

                line = line.rstrip('\n\r')

                if lineno in skip_linenos:
                    continue

                elif tablename_lineno == lineno:

                    table.set_tablename(line)

                elif options_lineno == lineno:

                    try:
                        parsed = csv.reader( [ line ] ).next()

                        for option_pair in parsed:

                            # If its a name value pair, its an option (otherwise its just a version number which we ignore)

                            m = re.search(r'^(\S+)=(\S+)$', option_pair, flags=re.IGNORECASE)
                            if m != None:
                                if m.group(1) == 'static' and self.istrue(m.group(2)):
                                    table.set_property('static', 1)
                                elif m.group(1) == 'update' and self.istrue(m.group(2)):
                                    table.set_property('update', 1)
                                elif m.group(1) == 'persist' and self.istrue(m.group(2)):
                                    table.set_property('persist', 1)
                                elif m.group(1) == 'keycols':
                                    table.set_property('keycols', m.group(2))
                                elif m.group(1) == 'identity':
                                    table.set_property('identity', m.group(2))

                    except Exception as inst:
                        raise Exception("Unable to parse options: %s" % (lineno + 1, inst))

                elif types_lineno == lineno:

                    try:
                        parsed = csv.reader( [ line ] ).next()
                        column_types = [ self.csv_to_qasino_type_map[x.strip()] for x in parsed ]
                    except Exception as inst:
                        raise Exception("Unsupported type in type list '%s' or parse error" % inst)
    
                    if column_names != None and len(column_names) != len(column_types):
                        raise Exception("Number of type names does not match number of column names! (line %d)" % lineno + 1)

                    table.set_column_types(column_types)

                elif colnames_lineno == lineno:

                    column_names = csv.reader( [ line ] ).next()
                
                    if column_types != None and len(column_names) != len(column_types):
                        raise Exception("Number of column names does not match number of column types! (line %d)" % lineno + 1)
    
                    table.set_column_names(column_names)

                # Data
                else:

                    input_row = csv.reader( [ line ] ).next()

                    ## Parse all the data into rows.
    
                    output_row = list()
    
                    # Read each column in data.

                    for column_index, column_cell in enumerate(input_row):
    
                        ##print "%d CELL: %s  name: %s" % (column_index, column_cell, column_names[column_index])
    
                        # Get the type so we know if it should be an int or not
    
                        try:
                            column_type = column_types[column_index]
                        except:
                            raise Exception("Could not find type for column %d on line %d!" % (column_index, lineno + 1))

                        try:
                            # Accept "null" fields as is (regardless of if they are an int/real etc)
                            if column_cell is None or column_cell == '':
                                output_row.append(column_cell)
                            elif column_type == 'INTEGER':
                                output_row.append(int(column_cell))
                            elif column_type == 'REAL':
                                output_row.append(float(column_cell))
                            else:
                                output_row.append(column_cell)
                        except Exception as e:
                            raise Exception("Parse error on line %d: %s" % (lineno + 1, str(e) ))
    
                    if table.add_row(output_row) == -1:
                        raise Exception("Wrong number of rows on line %d: '%s'" % (lineno + 1, line))
    
                # END if .. else .. 

            # END for each line

        except Exception as inst:
            #raise Exception("Csv read error on line %d" % lineno)
            logging.error('Csv read error on line %d: %s', lineno + 1, inst)
            return (None, str(inst))

        table.set_column_names(column_names)
        table.set_column_types(column_types)

        return ( table, None )



# Copyright (C) 2014 MediaMath, Inc. <http://www.mediamath.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import sqlite3
import re
import logging
import time
import string

from util import Identity, unicode_safe_str
import apsw_connection
import qasino_table

import apsw
from twisted.enterprise import adbapi
from twisted.internet import reactor

import thread

class CreateTableException(Exception):
    pass

class SqlConnections(object):
    """
    A pool of connections to a sql backend.
    Will be called either "sql_backend_writer" or "sql_backend_reader"
    by the data manager.
    """

    WRITER_INTERACTION = 0
    READER_INTERACTION = 1

    def __init__(self, filename, data_manager, archive_db_dir, thread_id, static_filename):
        self.data_manager = data_manager
        self.tables = {}
        self.connections = {}
        self.stats = {}

        self.archive_db_dir = archive_db_dir
        self.main_thread = thread_id
        self.open_new_db(filename, static_filename)

    def open_new_db(self, filename, static_filename):
        self.filename = filename
        self.static_filename = static_filename

        ##adbapi.noisy = True  # and add log.startLogging(sys.stdout) somewhere

        # since we're using apsw which is not technically DBAPI compliant 
        # I've created this wrapper class to make it act like one.
        adbapi.connectionFactory = apsw_connection.ApswConnection

        # if there is a static db ...
        def attach_static_db(conn):
            if static_filename != None and os.path.exists(static_filename):
                cur = conn.cursor()
                try:
                    result = cur.execute( "ATTACH DATABASE '%s' AS static;" % static_filename)
                except Exception as e:
                    logging.info("SqlConnections: ERROR Failed to attach static db '%s': %s", static_filename, str(e))

        # Open a connection pool for writing (one thread)...

        self.writer_dbpool = adbapi.ConnectionPool('apsw_connection', filename, 
                                                   cp_openfun=attach_static_db,
                                                   cp_min=1, cp_max=1)
                                                   
        self.writer_dbpool.connectionFactory = apsw_connection.ApswConnection

        # ...and one for reading (multiple threads).

        self.reader_dbpool = adbapi.ConnectionPool('apsw_connection', filename, 
                                                   cp_openfun=attach_static_db)

        self.reader_dbpool.connectionFactory = apsw_connection.ApswConnection


        # Preload list of tables already in db.

        self.reader_dbpool.runInteraction(self.preload_tables_list)
       

    def __del__(self):
        # This hack deals with when the condition where the last
        # reference to the sqlconnections object is held by a callback
        # and so the dtor is called from within a adbapi thread in the
        # thread pool.  If we are in one of these threads and call
        # dbpool.close it will throw an exception saying it can't join
        # the current thread - leaving a connection.  This may be a
        # non-issue and it only occurs with the first db created
        # (generation == 0) but all the same, detect when we are in a
        # different thread then the one we were contructed in and
        # schedule a call to shutdown (which must be static since this
        # is the dtor!) for later.

        if self.main_thread == thread.get_ident():
            SqlConnections.shutdown(self.writer_dbpool, self.filename, self.archive_db_dir)
            SqlConnections.shutdown(self.reader_dbpool, self.filename, self.archive_db_dir)
        else:
            reactor.callLater(0.1, SqlConnections.shutdown, self.writer_dbpool, self.filename, self.archive_db_dir)
            reactor.callLater(0.1, SqlConnections.shutdown, self.reader_dbpool, self.filename, self.archive_db_dir)

    @staticmethod
    def shutdown(dbpool, filename, archive_db_dir):
        """
        Facilitates cleaning up filesystem objects after the last
        user of the connection pool goes away.
        """

        try:
            dbpool.close()
        except Exception as e:
            logging.info("SqlConnections: dbpool.close failed in shutdown: %s", e)

        if filename != ':memory:':

            if archive_db_dir:
                logging.info("SqlConnections: Archiving db file '%s' to '%s'", filename, archive_db_dir)
                try:
                    os.renames(filename, archive_db_dir)
                except Exception as e:
                    logging.info("SqlConnections: ERROR: Archive failed! %s", e)
            else:
                logging.info("SqlConnections: Removing db file '%s'", filename)
                try:
                    os.remove(filename)
                except Exception as e:
                    pass  # ignore these
                    ##logging.info("SqlConnections: ERROR: Could not remove db file! %s", e)

    def do_select(self, txn, sql):

        self.stats['sql_received'] = self.stats.get('sql_received', 0) + 1

        try: 
            retval = txn.execute(sql)

        except Exception as e:

            self.stats['sql_errors'] = self.stats.get('sql_errors', 0) + 1

            return { "retval": 1, "error_message" : e }

        # Find the max column with for each column.

        max_widths = {}

        # Get the column names.

        column_names = []

        try:
            for column_index, column_name in enumerate(txn.getdescription()):

                if not max_widths.has_key(str(column_index)) or max_widths[str(column_index)] < len(column_name[0]):
                    max_widths[str(column_index)] = len(column_name[0])
            
                column_names.append(column_name[0])

        except apsw.ExecutionCompleteError as e:
            # Ugh, getdescription fails if the query succeeds but returns no rows!  This is the message:
            #    "Can't get description for statements that have completed execution"

            # For now return a zero row empty table:

            data = { "column_names" : [ "query returned zero rows" ], "rows" : [ ] }

            return { "retval" : 0, "error_message" : '', "data" : data, "max_widths" : { "0" : 24 } }

        # Save the data here:

        saved_rows = []
        nr_rows = 0

        # For each row.

        for row in txn.fetchall():

            saved_row = []

            # For each column.

            for column_index, cell in enumerate(row):

                # Convert all to strings and compute the length.

                cell = unicode_safe_str(cell)

                length = len(cell)

                if not max_widths.has_key(str(column_index)) or max_widths[str(column_index)] < length:
                    max_widths[str(column_index)] = length

                # Save the cell to the row for output later.

                saved_row.append(cell)

            # Save the row to the rows for output later.

            saved_rows.append(saved_row)


        self.stats['sql_completed'] = self.stats.get('sql_completed', 0) + 1

        data = { "column_names" : column_names,
                 "rows" : saved_rows }

        return { "retval" : 0, "error_message" : '', "data" : data, "max_widths" : max_widths }

    def get_schema(self, txn, table):

        schema = []

        try:
            txn.execute("pragma table_info( '%s' );" % table)
        except Exception as e:
            logging.info("SqlConnections: Failed to get table_info for '%s'.", table)
            return schema

        for row in txn.fetchall():
            schema.append( [ str(row[1]), str(row[2]) ] )

        return schema


    def do_sql(self, txn, sql):
        """ Execute generic sql """
        return txn.execute(sql)


    def do_desc(self, txn, tablename):

        schema = self.get_schema(txn, tablename)

        if schema == None or len(schema) <= 0:
            return (1, "Table %s not found" % (tablename,), None)

        table = { "column_names" : [ "column_name", "column_type" ],
                  "rows" : schema }

        return (0, '', table)
        
    def do_update_table(self, txn, table, identity):

        self.stats['updates_received'] = self.stats.get('updates_received', 0) + 1

        key_cols_str = table.get_property('keycols')

        column_names = table.get_column_names()

        if key_cols_str == None:
            key_column_names = column_names
        else:
            key_column_names = string.split(key_cols_str, ';')

        tablename = table.get_tablename()
        index_name = tablename + "_unique_index_" + '_'.join(key_column_names)

        create_index_sql = "CREATE UNIQUE INDEX IF NOT EXISTS %s ON %s (%s)" % (index_name, tablename, ', '.join(key_column_names))
        #logging.info("DEBUG: would execute: %s", create_index_sql)

        try:
            txn.execute(create_index_sql)

        except Exception as e:
            logging.info("ERROR: Failed to create unique index for table update: '%s': ( %s )", str(e), create_index_sql)

            self.stats['update_errors'] = self.stats.get('update_errors', 0) + 1
            return

        sql = "INSERT OR REPLACE INTO %s (%s) VALUES (%s)" % (tablename, ', '.join(column_names), ', '.join([ '?' for x in column_names ]) )

        rowcount = 0

        for row in table.get_rows():

            #logging.info("DEBUG: would execute: %s with %s", sql, row)
            txn.execute(sql, row)

            rowcount += txn.getconnection().changes()

        self.stats['updates_completed'] = self.stats.get('updates_completed', 0) + 1

        return rowcount

    def do_insert_table(self, txn, table):
        """
        Insert table into the backend.  Use multi-row insert
        statements (supported in sqlite 3.7.11) for speed.
        """
        
        self.stats['inserts_received'] = self.stats.get('inserts_received', 0) + 1

        rowcount = 0

        base_query = "INSERT INTO %s (%s) VALUES " % (table.get_tablename(), ",".join( table.get_column_names() ) )

        nr_columns = len(table.get_column_names())
        max_nr_values = 400  # supposedly the max is 500..

        nr_values = 0
        bind_values = []

        one_row = '({})'.format(','.join(['?'] * nr_columns))

        sql = base_query

        for row_index, row in enumerate(table.get_rows()):

            if nr_values + nr_columns > max_nr_values:

                # execute this batch and start again.

                sql += ','.join([ one_row ] * (nr_values / nr_columns))

                #logging.info("DEBUG: hit limit: executing (values %d, rows %d): %s", nr_values, rowcount, sql)

                txn.execute(sql, bind_values)
                bind_values = []
                nr_values = 0
                sql = base_query

            nr_values += nr_columns

            # add this row to our bind values

            for value in row:
                bind_values.append(value)

            rowcount += 1

        # handle last batch
        if rowcount > 0:
            sql += ','.join([ one_row ] * (nr_values / nr_columns))

            #logging.info("DEBUG: final: executing (values %d, rows %d): %s", nr_values, rowcount, sql)

            txn.execute(sql, bind_values)

        self.stats['inserts_completed'] = self.stats.get('inserts_completed', 0) + 1

        return rowcount

    def run_interaction(self, interaction_type, callback, *args, **kwargs):
        """ 
        Initiate a adbapi async interaction run a generic function, returns the deferred obj.
        """
        if interaction_type == SqlConnections.WRITER_INTERACTION:
            return self.writer_dbpool.runInteraction(callback, *args, **kwargs)

        elif interaction_type == SqlConnections.READER_INTERACTION:
            return self.reader_dbpool.runInteraction(callback, *args, **kwargs)

        return

    def insert_info_table(self, txn, db_generation_number, generation_start_epoch, generation_duration_s):
        """ 
        Adds a status table (qasino_server_info) to the database in each generation.
        """
        table = qasino_table.QasinoTable("qasino_server_info")
        table.add_column("generation_number",      "int")
        table.add_column("generation_duration_s",  "int")
        table.add_column("generation_start_epoch", "int")

        table.add_row( [ str(db_generation_number), generation_duration_s, generation_start_epoch ] )

        return self.add_table_data(txn, table, Identity.get_identity())

    def insert_sql_stats_table(self, txn, sql_backend_reader):
        """ 
        Adds a status table (qasino_server_sql_stats) to the database in each generation.
        Note we are actually saving stats from the "reader" backend because that is where 
        sql stats are logged.
        """
        table = qasino_table.QasinoTable("qasino_server_sql_stats")
        table.add_column("sql_received",  "int")
        table.add_column("sql_completed", "int")
        table.add_column("sql_errors",    "int")

        table.add_row( [ sql_backend_reader.stats.get('sql_received', 0),
                         sql_backend_reader.stats.get('sql_completed', 0),
                         sql_backend_reader.stats.get('sql_errors', 0) ] )

        return self.add_table_data(txn, table, Identity.get_identity())

    def insert_update_stats_table(self, txn):
        """ 
        Adds a status table (qasino_server_update_stats) to the database in each generation.
        """
        table = qasino_table.QasinoTable("qasino_server_update_stats")
        table.add_column("updates_received",  "int")
        table.add_column("updates_completed", "int")
        table.add_column("update_errors",     "int")
        table.add_column("inserts_received",  "int")
        table.add_column("inserts_completed", "int")

        table.add_row( [ self.stats.get('updates_received', 0), 
                         self.stats.get('updsates_completed', 0),
                         self.stats.get('update_errors', 0),
                         self.stats.get('inserts_received', 0), 
                         self.stats.get('inserts_completed', 0) ] )

        return self.add_table_data(txn, table, Identity.get_identity())

    def add_tables_table_rows(self, table):

        for tablename, table_data in self.tables.items():
            
            table.add_row( [ tablename, 
                             str(table_data["nr_rows"]),
                             str(table_data["updates"]),
                             table_data["last_update_epoch"],
                             0 if self.static_filename != None else 1 ]
                           )

    def preload_tables_list(self, txn):
        """ 
        This is used upon initial open of the db to get an actual list
        of tables from the db.
        """

        self.do_sql(txn, "SELECT tbl_name FROM sqlite_master WHERE type = 'table' and tbl_name NOT LIKE 'sqlite_%'")

        rows = txn.fetchall()

        for row in rows:

            tablename = row[0]

            try:
                self.do_sql(txn, "SELECT count(*) FROM %s;" % tablename)

                count = txn.fetchall()

                self.update_table_stats(tablename, count[0][0] if count[0][0] != None and count[0][0] > 0 else -1)
            except:
                logging.info("SqlConnections: Failed to get number of rows for table '%s': %s", tablename, str(e))
                pass

    def insert_connections_table(self, txn):
        """ 
        Adds a table (qasino_server_connections) to the database with per table info.
        """

        table = qasino_table.QasinoTable("qasino_server_connections")
        table.add_column("identity",          "varchar")
        table.add_column("nr_tables",         "int")
        table.add_column("last_update_epoch", "int")

        for connection, connection_data in self.connections.items():
            
            table.add_row( [ connection, # identity
                             str(len(connection_data["tables"])),
                             connection_data["last_update_epoch"] ] )

        return self.add_table_data(txn, table, Identity.get_identity())

    def add_views(self, txn, views):
        """
        Add all views to the backend.  This should not take a long time.
        """
        for viewname, viewdata in views.iteritems():

            logging.info("SqlConnections: Adding view '%s'", viewname)

            try:
                self.do_sql(txn, viewdata['view'])
                viewdata['loaded'] = True
                viewdata['error'] = ''
            except Exception as e:
                logging.info("SqlConnections: ERROR: Failed to add view '%s': %s", viewname, str(e))
                viewdata['loaded'] = False
                viewdata['error'] = str(e)

    def insert_views_table(self, txn, views):
        """ 
        Adds a table (qasino_server_connections) to the database with per table info.
        """

        table = qasino_table.QasinoTable("qasino_server_views")
        table.add_column("viewname", "varchar")
        table.add_column("loaded",   "int")
        table.add_column("errormsg", "varchar")
        table.add_column("view",     "varchar")

        for viewname, viewdata in views.iteritems():
            table.add_row( [ viewname, str(int(viewdata['loaded'])), str(viewdata['error']), viewdata['view'] ] )

        return self.add_table_data(txn, table, Identity.get_identity())

    def update_table_stats(self, tablename, nr_rows, identity=Identity.get_identity(), now=time.time(), sum=False):

        # Keep track of how many updates a table has received.

        if tablename not in self.tables:
            self.tables[tablename] = { "updates" : 1, 
                                       "nr_rows" : nr_rows,
                                       "last_update_epoch" : now }
        else:
            if sum:
                self.tables[tablename]["nr_rows"] += nr_rows

            self.tables[tablename]["updates"] += 1
            self.tables[tablename]["last_update_epoch"] = now


        # Keep track of which "identities" have added to a table.

        if identity not in self.connections:

            self.connections[identity] = { 'tables' : { tablename : nr_rows }, 
                                           'last_update_epoch' : now }
        else:
            self.connections[identity]["last_update_epoch"] = now

            if sum and tablename in self.connections[identity]["tables"]:
                self.connections[identity]["tables"][tablename] += nr_rows
            else:
                self.connections[identity]["tables"][tablename] = nr_rows

    def async_add_table_data(self, *args, **kwargs):
        """
        Initiate a adbapi async interaction to add a table to the backend, returns the deferred obj.
        """
        return self.writer_dbpool.runInteraction(self.add_table_data, *args, **kwargs)

    def add_table_data(self, txn, table, identity):
        """ 
        Add a table to the backend if it doesn't exist and insert the data.
        This is executed using adbapi in a thread pool.
        """

        start_time = time.time()

        # For when we hit lock contention - only retry so many times.

        table.init_retry(5)

        tablename = table.get_tablename()

        if tablename is None:
            logging.info("SqlConnections: Unknown tablename from '%s'", identity)
            return

        update = table.get_property('update')
        static = table.get_property('static')
        persist = table.get_property('persist')

        # Check if we need to save the table in case of persistence.

        self.data_manager.check_save_table(table, identity)

        # Now check if we've already added the table.

        try:
            if not update and not static and self.connections[identity]["tables"][tablename] > 0:
                logging.info("SqlConnections: '%s' has already sent an update or is persisted for table '%s'", identity, tablename)
                return
        except KeyError:
            # Ignore
            pass

        try:
            # Wrap the create, alter tables and inserts or updates in
            # a transaction... All or none please.

            txn.execute("BEGIN;")

            # Use this to detect if the table already exists - we
            # could just not do this and use the create table to
            # detect if a table is already there or not but a error on
            # the create table would mean having to restart the
            # transaction and require some weird retry logic..

            create_sql = "CREATE TABLE IF NOT EXISTS '%s' ( \n" % tablename

            columns = []
            
            for column_name, column_type in table.zip_columns():
                columns.append( "'%s' %s DEFAULT NULL" % (column_name, column_type) )

            create_sql += ",\n".join( columns )

            create_sql += ")"

            try:
                result = self.do_sql(txn, create_sql)
            except Exception as e:
                errorstr = str(e)

                # Is this a "table already exists" error?  Then its a race condition between checking if a table exists and actually creating it.
                # This should never happen now - we are using CREATE IF NOT EXISTS ...
                if "already exists" in errorstr:
                    # Don't need to log this as it is fairly common.
                    ##logging.info("SqlConnections: Warning: Create table race condition '%s' for %s: '%s' ... retrying", tablename, identity, errorstr)
                    raise CreateTableException("Create table race condition '%s' for %s: '%s'" % (tablename, identity, errorstr,))
                
                else:
                    # Otherwise some kind of sql error, pass the exception up
                    raise e


            # Merge the schemas of what might be already there and our update.
            schema = self.get_schema(txn, tablename)

            self.data_manager.table_merger.merge_table(txn, table, schema, self)

            # Now update the table.
            # Truncate and re-insert if this is static.
            # Call do_update if this is an update type.
            # Otherwise we insert.

            if static:
                txn.execute("DELETE FROM '%s';" % tablename)
                rowcount = self.do_insert_table(txn, table)

            else:
                if update:
                    rowcount = self.do_update_table(txn, table, identity)
                else:
                    rowcount = self.do_insert_table(txn, table)

            txn.execute("COMMIT;")

        except apsw.BusyError as e:
            # This should only happen if we exceed the busy timeout.
            logging.info("SqlConnections: Warning: Lock contention for adding table '%s': %s for %s", tablename, str(e), identity)

            txn.execute("ROLLBACK;")

            # Should we try again?
            if table.test_retry():
                logging.info("SqlConnections: ERROR: Retries failed for table '%s'", tablename)
                return 0

            # Place this request back on the event queue for a retry.
            self.async_add_table_data(table, identity)

            return 0

        except CreateTableException as e:
            # Message when raised

            txn.execute("ROLLBACK;")

            # Should we try again?
            if table.test_retry():
                logging.info("SqlConnections: ERROR: Retries failed for table '%s'", tablename)
                return 0

            # Place this request back on the event queue for a retry.
            self.async_add_table_data(table, identity)

            return 0

        except Exception as e:
            logging.info("SqlConnections: ERROR: Failed to add table '%s': %s", tablename, e)

            txn.execute("ROLLBACK;")

            return 0

        now = time.time()

        nr_rows = table.get_nr_rows()

        properties_str = ''
        if static: properties_str += ' static'
        if update: properties_str += ' update'
        if persist: properties_str += ' persist'

        logging.info("SqlConnections: New data%s for table '%s' %d rows from '%s' (%.02f seconds)", 
                     properties_str, tablename, rowcount, identity, now - start_time)

        # Update some informational stats used for making the tables
        # and connections tables.

        self.update_table_stats(tablename, rowcount, identity=identity,
                                sum=not update and not static, now=now)
            
        return 1






# Copyright (C) 2014 MediaMath, Inc. <http://www.mediamath.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from txzmq import ZmqFactory, ZmqEndpoint, ZmqEndpointType, ZmqREQConnection

import logging

import json

from util import Identity

class ZmqRequestor(ZmqREQConnection):

    def __init__(self, remote_host, port, zmq_factory, data_manager=None):

        self.data_manager = data_manager

        self.remote_host = remote_host

        endpoint = ZmqEndpoint(ZmqEndpointType.connect, "tcp://%s:%d" % (remote_host, port))

        ZmqREQConnection.__init__(self, zmq_factory, endpoint)

    def request_metadata(self):
        msg = { "op" : "get_table_list", "identity" : Identity.get_identity() }
        #logging.info("ZmqRequestor: Requesting table list from %s.", self.remote_host)
        deferred = self.sendMsg(json.dumps(msg))
        deferred.callback = self.message_received

    def send_table(self, table):
        deferred = self.sendMsg(table.get_json(op="add_table_data", identity=Identity.get_identity()))
        deferred.callback = self.message_received
        

    def message_received(self, msg):
        response_meta = json.loads(msg[0])

        if response_meta == None or response_meta["response_op"] == None:
            logging.error("ZmqRequestor: bad message response received")
        elif response_meta["response_op"] == "tables_list":
            logging.info("ZmqRequestor: Table list response: %s", json.loads(msg[1]))
        elif response_meta["response_op"] == "ok":
            logging.info("ZmqRequestor: request OK")
        elif response_meta["response_op"] == "error":
            logging.info("ZmqRequestor: request ERROR: " + response_meta["error_message"])
        else:
            logging.error("ZmqRequestor: unknown response: ", response_meta)



# Copyright (C) 2014 MediaMath, Inc. <http://www.mediamath.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import apsw
import thread

def connect(file):
    conn = apsw.Connection(file)
    conn.setbusytimeout(1000) # 1 second
    return conn
    
class ApswConnection(object):
    """
    A wrapper for a apsw connection instance.

    The wrapper passes almost everything to the wrapped connection and so has
    the same API. However, the Connection knows about its pool and also
    handle reconnecting should when the real connection dies.
    """

    def __init__(self, pool):
        self._pool = pool
        self._connection = None
        self.reconnect()

    def close(self):
        # The way adbapi works right now means that closing a connection is
        # a really bad thing  as it leaves a dead connection associated with
        # a thread in the thread pool.
        # Really, I think closing a pooled connection should return it to the
        # pool but that's handled by the runWithConnection method already so,
        # rather than upsetting anyone by raising an exception, let's ignore
        # the request
        pass

    def rollback(self):
        # Do not call commit or rollback because they don't exist:
        # http://apidoc.apsw.googlecode.com/hg/dbapi.html#connection-objects
        # The user of this class will just execute BEGIN, COMMIT, ROLLBACK as needed.
        return

    def commit(self):
        # Do not call commit or rollback because they don't exist:
        # http://apidoc.apsw.googlecode.com/hg/dbapi.html#connection-objects
        # The user of this class will just execute BEGIN, COMMIT, ROLLBACK as needed.
        return

    def reconnect(self):
        if self._connection is not None:
            self._pool.disconnect(self._connection)
        self._connection = self._pool.connect()

    def __getattr__(self, name):
        return getattr(self._connection, name)

__all__ = ['ApswConnection']



# Copyright (C) 2014 MediaMath, Inc. <http://www.mediamath.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import json
import re
import time
import sqlite3
import StringIO

from pprint import pprint
from twisted.web.resource import Resource
from twisted.web.server import NOT_DONE_YET
from twisted.web import http

import util
import qasino_table
import csv_table_reader

from jinja2 import Environment, FileSystemLoader

class UIResource(Resource):

    isLeaf = True

    def __init__(self, templates_dir, data_manager):
        self.data_manager = data_manager
        self.jinja = Environment(loader=FileSystemLoader(templates_dir))

    def render_page(self, page_name, request, context={}):
        
        request.setHeader("Content-Type", "text/html; charset=utf-8")

        ctx = { 'page' : page_name,
                'params' : request.args }

        # Do it this way so input can completely override.
        for k, v in context.iteritems():
            ctx[k] = v

        template = self.jinja.get_template(page_name + ".html.j2")

        output = template.render(ctx)

        return output.encode('utf-8')


    def render_page_with_sql(self, page_name, sql, request, context={}):

        request.setHeader("Content-Type", "text/html; charset=utf-8")

        query_id = self.data_manager.get_query_id()
        query_start = time.time()

        d = self.data_manager.async_validate_and_route_query(sql, query_id)

        d.addCallback(self.sql_complete_callback, page_name, query_id, query_start, request, context=context)

        return NOT_DONE_YET
    
    def sql_complete_callback(self, result, page_name, query_id, query_start, request, context={}):

        data = result["data"] if "data" in result else None

        ctx = { 'page' : page_name,
                'rows' : [],
                'column_names' : [],
                'params' : request.args,
                }

        # Do it this way so input can completely override.
        for k, v in context.iteritems():
            ctx[k] = v


        template = self.jinja.get_template(page_name + ".html.j2")

        err_msg = str(result.get('error_message', ''))

        if data is None and err_msg == '':

            ctx['error_message'] = 'No data!'

        elif err_msg != '':

            ctx['error_message'] = err_msg

        elif "rows" in data and "column_names" in data:

            ctx['rows'] = data['rows']
            ctx['column_names'] = data['column_names']

        else:
            ctx['error_message'] = 'No row or column data in result!'

        output = template.render(ctx)
        request.write( output.encode('utf-8') )

        request.finish()


class UIResourceTables(UIResource):

    def render_GET(self, request):

        sql = "select tablename, nr_rows, nr_updates, datetime(last_update_epoch, 'unixepoch') last_update, static from qasino_server_tables order by tablename;"

        name_map = { 'tablename' : 'Tablename',
                     'nr_rows' : 'Nr Rows',
                     'nr_updates' : 'Nr Updates',
                     'last_update' : 'Last Update',
                     'static' : 'Static',
                     }

        return self.render_page_with_sql("tables", sql, request, context={ 'pretty_column_names_map' : name_map,
                                                                           'links' : { 'tablename' : '/desc?tablename=%s' } })


class UIResourceDesc(UIResource):

    def render_GET(self, request):

        if 'tablename' in request.args:
            
            sql = "desc {};".format(request.args['tablename'][0].replace(" ", ""))

            name_map = { 'column_name' : 'Column Name',
                         'column_type' : 'Column Type',
                         }

            return self.render_page_with_sql("desc", sql, request, context={ 'pretty_column_names_map' : name_map })
        else:

            return self.render_page("desc", request)


class UIResourceQuery(UIResource):

    def render_GET(self, request):

        if 'sql' in request.args:
            
            m = re.search(r"^\s*select\s+", request.args['sql'][0], flags=re.IGNORECASE)
            if m == None:
                
                return self.render_page("query", request, 
                                        context={ 'error_message' : 'Please use only SELECT statements.' })

            sql = "{};".format(request.args['sql'][0])

            return self.render_page_with_sql("query", sql, request)
        else:

            return self.render_page("query", request)




#!/usr/bin/python

# Copyright (C) 2014 MediaMath, Inc. <http://www.mediamath.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import signal
import cmd
import readline
import json
import re
import sqlite3
from optparse import OptionParser
import zmq
import requests
import getpass

for path in [
    os.path.join('opt', 'qasino', 'lib'),
    os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'lib'))
]:
    if os.path.exists(os.path.join(path, '__init__.py')):
        sys.path.append(path)
        break

import zmq_requestor
import constants
import util

class QasinoZMQConnection(object):
    def __init__(self, options):
        self.options = options
        self.socket = None

    def connect(self):
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.REQ)
        self.socket.connect("tcp://%s:%s" % (self.options.hostname, self.options.port))

    def send_and_get_response(self, json):
        if self.socket == None:
            return None
        self.socket.send(json)
        return self.socket.recv()


class QasinoHttpConnection(object):
    def __init__(self, options):
        self.options = options

    def connect(self):
        self.requests_conn = requests.Session()

    def send_and_get_response(self, json):
        if self.requests_conn == None:
            return None

        URL = 'https://%s:%d/request?op=query' % (self.options.hostname, self.options.port)

        request_options = { 'headers' : {'Content-Type': 'application/json'},
                            'data' : json } 

        if self.options.skip_ssl_verify:
            request_options['verify'] = False
        if self.options.username and self.options.password:
            request_options['auth'] = (self.options.username, self.options.password)

        try:
            response = self.requests_conn.post(URL, **request_options)
        except Exception as e:
            print "ERROR: failed to send HTTPS POST to '%s': %s" % (URL, str(e))
            return None

        return response.text

class QasinoCmd(cmd.Cmd):

    multiline_prompt = '      > '
    default_prompt   = 'qasino> '
    prompt = default_prompt

    sql_statement = ''

    def __init__(self, conn):
        self.conn = conn
        self.use_write_db = False
        cmd.Cmd.__init__(self)

    def send_query(self, sql_statement, use_write_db=False):

        request_meta = { "op" : "query", "sql" : sql_statement,
                         "use_write_db" : use_write_db }

        response = self.conn.send_and_get_response(json.dumps(request_meta))
        if response == None:
            print "ERROR: error communicating with server"
            return

        try:
            obj = json.loads(response)
        except Exception as e:
            print "ERROR: unable to parse response from server: %s: %s" % (response, str(e))
            return

        if obj == None or "response_op" not in obj:
            print "ERROR: invalid response from server: ", obj

        elif obj["response_op"] == "error":
            if "error_message" in obj:
                print "ERROR: %s" % str(obj["error_message"])
            else:
                print "ERROR: Unknown error from server."

        elif obj["response_op"] == "result_table":

            max_widths = obj["max_widths"] if "max_widths" in obj else None

            util.pretty_print_table(self, obj["table"], max_widths=max_widths)
            #print "Got response: ", obj

        else:
            print "ERROR: Unknown op response from server."

    def sendLine(self, line):
        print line


    def is_set_value_true(self, value):
        m = re.search(r'^\s*(true|1|yes)\s*$', value, flags=re.IGNORECASE)
        if m:
            return 1

        m = re.search(r'^\s*(false|0|no)\s*$', value, flags=re.IGNORECASE)
        if m:
            return 0

        return -1

    def set_set(self, name, value):

        if name == 'use_write_db':
            truth = self.is_set_value_true(value)
            if truth == 1:
                print "Using write database."
                self.use_write_db = True
            elif truth == 0:
                print "Using read database."
                self.use_write_db = False
            else:
                print "Invalid value for '%s'." % (name,)
        else:
            print "Unrecognized SET statement: %s = %s" % (name, value)


    def default(self, line):
        self.sql_statement = self.sql_statement + ' ' + line.rstrip('\n')

        if sqlite3.complete_statement(self.sql_statement):

            match_set = re.search(r"^\s*set\s+(\S+)\s*=\s*'?(\S+?)'?\s*;\s*$", self.sql_statement, flags=re.IGNORECASE)
            if match_set:

                name = match_set.group(1)
                value = match_set.group(2)

                self.set_set(name, value)

            else:
                m = re.search(r"^\s*select\s+", self.sql_statement, flags=re.IGNORECASE)
                m2 = re.search(r"^\s*show\s+", self.sql_statement, flags=re.IGNORECASE)
                m3 = re.search(r"^\s*desc\s+", self.sql_statement, flags=re.IGNORECASE)

                if m or m2 or m3:
                    self.send_query(self.sql_statement, self.use_write_db)
                else:
                    print "Unrecognized statement: ", self.sql_statement

            self.reset_multiline()

        else:
            self.multiline()

    def multiline(self):
        self.prompt = self.multiline_prompt

    def reset_multiline(self):
        self.prompt = self.default_prompt
        self.sql_statement = ''
            
    def emptyline(self):
        pass

    def do_EOF(self, line):
        return True

def signal_handler(signum, frame):
    sig_names = dict((k, v) for v, k in signal.__dict__.iteritems() if v.startswith('SIG'))
    print "Caught %s.  Exiting..." % sig_names[signum]
    exit(0)

def run():

    parser = OptionParser()

    parser.add_option("-I", "--identity", dest="identity",
                      help="Use IDENTITY as identity", metavar="IDENTITY")

    parser.add_option("-H", "--hostname", dest="hostname", default='localhost',
                      help="Send table to HOSTNAME qasino server", metavar="HOSTNAME")

    parser.add_option("-p", "--port", dest="port", default=constants.HTTPS_PORT,
                      help="Use PORT for qasino server", metavar="PORT")

    parser.add_option("-t", "--use-https", dest="use_https", default=False,
                      action='store_true',
                      help="Use HTTPS transport for making queries")

    parser.add_option("-z", "--use-zmq", dest="use_zmq", default=False,
                      action='store_true',
                      help="Use ZeroMQ transport for making queries")

    parser.add_option("-u", "--username", dest="username", 
                      help="HTTPS auth username")

    parser.add_option("-w", "--password", dest="password", 
                      help="HTTPS auth password")

    parser.add_option("-s", "--skip-ssl-verify", dest="skip_ssl_verify", default=False, 
                      action="store_true",
                      help="Don't verify SSL certificates.")

    parser.add_option("-q", "--query", dest="query", 
                      action="append",
                      help="Query to execute and exit.")

    (options, args) = parser.parse_args()

    if options.identity != None:
        Identity.set_identity(options.identity)

    if options.hostname == None:
        print "Please specify a hostname to connect to."
        exit(1)

    # Can't do both.
    if options.use_https and options.use_zmq:
        print "Can not use both ZMQ and HTTPS transports."
        exit(1)

    # Pick a default.
    if not options.use_https and not options.use_zmq:
        options.use_zmq = True

    # Catch signals

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    if options.use_zmq and options.port == constants.HTTPS_PORT:
        # switch to the right default port for zmq
        options.port = constants.ZMQ_RPC_PORT
    
    print "Connecting to %s %s:%s." % ('https' if options.use_https else 'zmq', 
                                       options.hostname, options.port)


    if options.use_https:

        if not options.username:
            options.username = os.environ.get('QASINO_SQLCLIENT_USERNAME')
            
        if not options.password:
            options.password = os.environ.get('QASINO_SQLCLIENT_PASSWORD')

        if options.username and not options.password:
            options.password = getpass.getpass()

        conn = QasinoHttpConnection(options)
        conn.connect()

    if options.use_zmq:
        # switch to the right default port
        if options.port == constants.HTTPS_PORT:
            options.port = constants.ZMQ_RPC_PORT

        conn = QasinoZMQConnection(options)
        conn.connect()

    if options.query:

        # Command line query string - execute and exit.
        for query in options.query:
            QasinoCmd(conn).send_query(query)

    else:

        # Enter the command loop.
        QasinoCmd(conn).cmdloop()
        
if __name__ == "__main__":
    run()




# Copyright (C) 2014 MediaMath, Inc. <http://www.mediamath.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket
import fcntl
import struct
import uuid
import random

import qasino_table

class Identity(object):
    identity = 'unidentified'

    @staticmethod
    def get_ip_address_from_iface(ifname):
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        return socket.inet_ntoa(
                        fcntl.ioctl(
                              s.fileno(),
                              0x8915,  # SIOCGIFADDR
                              struct.pack('256s', ifname[:15])
                        )[20:24]
               )

    @staticmethod
    def get_ip_address_from_hostname():
        return socket.gethostbyname(socket.gethostname())

    @staticmethod
    def get_identity():
        
        # cached or pre-set?

        if Identity.identity == 'unidentified':

            # python newbie here..
            try:
                ip = Identity.get_ip_address_from_iface('eth0')
            except:
                try: 
                    ip = Identity.get_ip_address_from_iface('eth1')
                except:
                    try:
                        ip = Identity.get_ip_address_from_hostname()
                    except:
                        ip = None

            if ip == None or ip.startswith('127.') or ip.startswith('192.168.') or ip.startswith('10.'):
                # Use a mac address if we didn't get an ip address
                Identity.identity = str(uuid.getnode())
            else:
                Identity.identity = ip

        return Identity.identity

    @staticmethod
    def set_identity(_identity):
        Identity.identity = _identity



def pretty_print_table(outputter, table, max_widths=None, column_delim="  "):
    """
    Expects table to be:

    table = { "column_names" : [ "col1", "col2", "col3" ],
              "rows" : [ [ "row1cell1", "row1cell2", "row1cell3" ],
                         [ "row2cell1", "row2cell2", "row2cell3" ]
                       ]
            }

    Outputs:
         col1       col2       col3
    =========  =========  =========
    row1cell1  row1cell2  row1cell3
    row2cell1  row2cell2  row2cell3
    """

    if max_widths == None:

        max_widths = {}

        # Find the max widths of all the columns.

        # First check the column names.

        for column_index, column_name in enumerate(table["column_names"]):

            length = len(column_name)

            if not max_widths.has_key(str(column_index)) or max_widths[str(column_index)] < length:
                max_widths[str(column_index)] = length
    
        # Now the data.

        for row in table["rows"]:

            for column_index, cell in enumerate(row):

                length = len(unicode_safe_str(cell))

                if not max_widths.has_key(str(column_index)) or max_widths[str(column_index)] < length:
                    max_widths[str(column_index)] = length
                

    # Print the column names

    outputter.sendLine(
        column_delim.join( [ str(column_name.rjust(max_widths[str(index)]))
                             for index, column_name in enumerate(table["column_names"]) 
                           ] 
                         )
    )
    
    # Print ==== under column names

    outputter.sendLine(
        column_delim.join( [ "=" * max_widths[str(index)]
                             for index, column_name in enumerate(table["column_names"]) 
                           ] 
                         )
    )

    nr_rows = 0

    # Print the rows now.

    for row in table["rows"]:

        nr_rows += 1

        line = column_delim.join( [ unicode_safe_str(cell).rjust(max_widths[str(index)]) 
                                    for index, cell in enumerate(row) 
                                  ] 
                                )

        outputter.sendLine(line)

    # Now number of rows

    outputter.sendLine("%d rows returned" % nr_rows)

def unicode_safe_str(s):
    if s is None:    # this is an added "feature"
        return ''
    elif type(s) is unicode:
        # I can't get jinja in the http ui to encode as utf-8 instead of ascii (which barfs) so escape for now.
        #return s.encode('utf-8')
        return s.encode('unicode_escape')
    else:
        return str(s)


def random_string(start, stop):
    string = ""
    sizeof_string = random.randint(start, stop + 1)
    for x in range(sizeof_string):
        pick_a_char_index = random.randint(0, len(random_string.alphabet) - 1)
        string += random_string.alphabet[pick_a_char_index]
    return string

random_string.alphabet = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"



# Below included from:
# http://code.activestate.com/recipes/325204-passwd-file-compatible-1-md5-crypt/
# -------------------------------------------------------------------------------

# Based on FreeBSD src/lib/libcrypt/crypt.c 1.2
# http://www.freebsd.org/cgi/cvsweb.cgi/~checkout~/src/lib/libcrypt/crypt.c?rev=1.2&content-type=text/plain

# Original license:
# * "THE BEER-WARE LICENSE" (Revision 42):
# * <phk@login.dknet.dk> wrote this file.  As long as you retain this notice you
# * can do whatever you want with this stuff. If we meet some day, and you think
# * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp

# This port adds no further stipulations.  I forfeit any copyright interest.

import md5

def md5crypt(password, salt, magic='$1$'):
    # /* The password first, since that is what is most unknown */ /* Then our magic string */ /* Then the raw salt */
    m = md5.new()
    m.update(password + magic + salt)

    # /* Then just as many characters of the MD5(pw,salt,pw) */
    mixin = md5.md5(password + salt + password).digest()
    for i in range(0, len(password)):
        m.update(mixin[i % 16])

    # /* Then something really weird... */
    # Also really broken, as far as I can tell.  -m
    i = len(password)
    while i:
        if i & 1:
            m.update('\x00')
        else:
            m.update(password[0])
        i >>= 1

    final = m.digest()

    # /* and now, just to make sure things don't run too fast */
    for i in range(1000):
        m2 = md5.md5()
        if i & 1:
            m2.update(password)
        else:
            m2.update(final)

        if i % 3:
            m2.update(salt)

        if i % 7:
            m2.update(password)

        if i & 1:
            m2.update(final)
        else:
            m2.update(password)

        final = m2.digest()

    # This is the bit that uses to64() in the original code.

    itoa64 = './0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'

    rearranged = ''
    for a, b, c in ((0, 6, 12), (1, 7, 13), (2, 8, 14), (3, 9, 15), (4, 10, 5)):
        v = ord(final[a]) << 16 | ord(final[b]) << 8 | ord(final[c])
        for i in range(4):
            rearranged += itoa64[v & 0x3f]; v >>= 6

    v = ord(final[11])
    for i in range(2):
        rearranged += itoa64[v & 0x3f]; v >>= 6

    return magic + salt + '$' + rearranged

#if __name__ == '__main__':
#
#    def test(clear_password, the_hash):
#        magic, salt = the_hash[1:].split('$')[:2]
#        magic = '$' + magic + '$'
#        return md5crypt(clear_password, salt, magic) == the_hash
#
#    test_cases = (
#        (' ', '$1$yiiZbNIH$YiCsHZjcTkYd31wkgW8JF.'),
#        ('pass', '$1$YeNsbWdH$wvOF8JdqsoiLix754LTW90'),
#        ('____fifteen____', '$1$s9lUWACI$Kk1jtIVVdmT01p0z3b/hw1'),
#        ('____sixteen_____', '$1$dL3xbVZI$kkgqhCanLdxODGq14g/tW1'),
#        ('____seventeen____', '$1$NaH5na7J$j7y8Iss0hcRbu3kzoJs5V.'),
#        ('__________thirty-three___________', '$1$HO7Q6vzJ$yGwp2wbL5D7eOVzOmxpsy.'),
#        ('apache', '$apr1$J.w5a/..$IW9y6DR0oO/ADuhlMF5/X1')
#    )
#
#    for clearpw, hashpw in test_cases:
#        if test(clearpw, hashpw):
#            print '%s: pass' % clearpw
#        else:
#            print '%s: FAIL' % clearpw

# We'll use it like this for our cmp_pass function in qasino_server.py
def get_apache_md5(clear_password, the_hash):
    magic, salt = the_hash[1:].split('$')[:2]
    magic = '$' + magic + '$'
    return md5crypt(clear_password, salt, magic)




# Copyright (C) 2014 MediaMath, Inc. <http://www.mediamath.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from txzmq import ZmqFactory, ZmqEndpoint, ZmqEndpointType, ZmqREPConnection
import logging
import json
import time
import sqlite3

from util import Identity
import qasino_table

class ZmqReceiver(ZmqREPConnection):

    def __init__(self, port, zmq_factory, data_manager):
        
        self.data_manager = data_manager

        endpoint = ZmqEndpoint(ZmqEndpointType.bind, "tcp://*:%d" % port)

        ZmqREPConnection.__init__(self, zmq_factory, endpoint)
                                  
    def gotMessage(self, messageId, *messageParts):

        try:
            obj = json.loads(messageParts[0])
        except Exception as e:
            logging.info("ZmqReceiver: ERROR failed to get/parse content of POST: %s", str(e))
            response_meta = { "response_op" : "error", "error_message" : "Failed to parse JSON message: %s" % str(e), "identity" : Identity.get_identity() }
            self.reply(messageId, json.dumps(response_meta))
            return

        response_meta = { "response_op" : "error", "identity" : Identity.get_identity(), "error_message" : "Unspecified error" }

        if obj == None or obj["op"] == None:
            logging.error("ZmqReceiver: Error, unrecognized message.")
            response_meta = { "response_op" : "error", "error_message" : "Unrecognized request", "identity" : Identity.get_identity() }
            self.reply(messageId, json.dumps(response_meta))

        elif obj["op"] == "get_table_list":
            #logging.info("ZmqReceiver: Got request for table list.")
            response_meta = { "response_op" : "tables_list", "identity" : Identity.get_identity() }
            response_data = self.data_manager.get_table_list()
            self.reply(messageId, json.dumps(response_meta), json.dumps(response_data))

        elif obj["op"] == "add_table_data":
            #logging.info("ZmqReceiver: Got request to add data.")
            #print "Got request: ", obj
            table = qasino_table.QasinoTable()
            err = table.from_obj(obj)
            if err is not None:
                errmsg = "Invalid input format: " + str(err)
                logging.info("ZmqReceiver: " + errmsg)
                response_meta = { "response_op" : "error", "identity" : Identity.get_identity(), "error_message" : errmsg }
            else:
                response_meta = { "response_op" : "ok", "identity" : Identity.get_identity() }
                try:
                    if table.get_property("static"):
                        self.data_manager.sql_backend_writer_static.async_add_table_data(table, table.get_property("identity"))
                    else:
                        self.data_manager.sql_backend_writer.async_add_table_data(table, table.get_property("identity"))
                except Exception as e:
                    response_meta = { "response_op" : "error", "identity" : util.Identity.get_identity(), "error_message" : str(e) }

            self.reply(messageId, json.dumps(response_meta))

        elif obj["op"] == "generation_signal":
            logging.info("ZmqReceiver: Got generation signal.")
            # Currently unused..
            
        elif obj["op"] == "query":
            #logging.info("ZmqReceiver: Got request for table list.")
            use_write_db = True if "use_write_db" in obj and obj["use_write_db"] else False
            if "sql" not in obj:
                response_meta = { "response_op" : "error", "error_message" : "Must specify sql", "identity" : Identity.get_identity() }
                self.reply(messageId, json.dumps(response_meta))
            else:
                try:
                    self.process_sql_statement(obj["sql"], messageId, use_write_db=use_write_db)
                except Exception as e:
                    logging.error('ZmqReceiver: Invalid message received from client: error="%s", msg="%s"', str(e), str(obj))
                    response_meta = { "response_op" : "error", "error_message" : str(e), "identity" : Identity.get_identity() }
                    self.reply(messageId, json.dumps(response_meta))

                # Response is handled in the callback.
                return

        else:
            logging.error("ZmqReceiver: Error, unrecognized op '%s'", obj["op"])
            response_meta = { "response_op" : "error", "identity" : Identity.get_identity(), "error_message" : "Unrecognized op '%s'" % obj["op"] }
            self.reply(messageId, json.dumps(response_meta))


    
    def process_sql_statement(self, sql_statement, messageId, use_write_db=False):

        query_id = self.data_manager.get_query_id()

        query_start = time.time()

        logging.info("ZmqReceiver: (%d) SQL received: %s", query_id, sql_statement.rstrip())

        if not sqlite3.complete_statement(sql_statement):

            # Try adding a semicolon at the end.
            sql_statement = sql_statement + ";"

            if not sqlite3.complete_statement(sql_statement):
                return self.sql_complete_callback( { "retval" : 1, 
                                                     "error_message" : "Incomplete sql statement" },
                                                   query_id, query_start )

            # else continue

        # Enqueue
        d = self.data_manager.async_validate_and_route_query(sql_statement, query_id, use_write_db=use_write_db)

        d.addCallback(self.sql_complete_callback, query_id, query_start, messageId)

        
    def sql_complete_callback(self, result, query_id, query_start, messageId):
        """
        Called when a sql statement completes.
        """

        # To start just our identity.
        response_meta = { "identity" : Identity.get_identity() }

        retval = result["retval"]
        error_message = ''
        if "error_message" in result:
                error_message = str(result["error_message"])

        # Success?

        if retval == 0 and "data" in result:

            response_meta["response_op"] = "result_table"
            response_meta["table"] = result["data"]

            if "max_widths" in result:
                response_meta["max_widths"] = result["max_widths"]

        # Or error?

        if retval != 0:
            logging.info("ZmqReceiver: (%d) SQL error: %s", query_id, error_message)

            response_meta["response_op"] = "error"
            response_meta["error_message"] = error_message

        else:
            logging.info("ZmqReceiver: (%d) SQL completed (%.02f seconds)", query_id, time.time() - query_start)

        # Send the response!

        self.reply(messageId, json.dumps(response_meta))



# Copyright (C) 2014 MediaMath, Inc. <http://www.mediamath.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import json
import re
import time
import sqlite3
import StringIO

from pprint import pprint
from twisted.web.resource import Resource
from twisted.web.server import NOT_DONE_YET
from twisted.web import http

import util
import qasino_table
import csv_table_reader

class MyLoggingHTTPChannel(http.HTTPChannel):
    def connectionMade(self):
        logging.info("HttpReceiver: Connection from '%s'.", str(self.transport.getPeer().host))
        http.HTTPChannel.connectionMade(self)

    def connectionLost(self, reason):
        logging.info("HttpReceiver: Connection close '%s'.", str(self.transport.getPeer().host))
        http.HTTPChannel.connectionLost(self, reason)

class HttpReceiver(Resource):

    def __init__(self, data_manager):
        self.data_manager = data_manager

    def render_GET(self, request):

        request.setHeader("Content-Type", "application/json")

        #pprint(request.__dict__)

        if 'op' not in request.args:
            logging.error("HttpReceiver: No op specified for GET request: %s", ','.join(request.args))
            response_meta = { "response_op" : "error", "identity" : util.Identity.get_identity(), "error_message" : "No op specified" }
            return json.dumps(response_meta)

        ## A simple name value update op.

        if request.args['op'][0] == "name_value_update":
            logging.info("HttpReceiver: Name value update.")

            if 'name' in request.args and 'value' in request.args:
                
                identity = request.args['identity'][0] if 'identity' in request.args else 'unknown'
                
                # Parse the name into '<tablename>.<column>'
                
                name = request.args['name'][0]

                m = re.search(r'^([\w_]+)\.([\w_]+)$', name)
                if m == None:
                    logging.info("HttpReciever: Invalid name in name value update: '%s'", name)
                    response_meta = { "response_op" : "error", "error_message" : "Invalid name in namve value update", "identity" : util.Identity.get_identity() }
                    return json.dumps(response_meta)

                tablename = m.group(1)
                columnname = m.group(2)
                value = request.args['value'][0]

                table = qasino_table.QasinoTable(tablename)
                table.add_column('identity', 'varchar')
                table.add_column(columnname, 'varchar')
                table.add_row( [ identity, value ] )
                table.set_property('update', 1)
                table.set_property('persist', 1)
                table.set_property('keycols', 'identity')

                self.data_manager.sql_backend_writer.async_add_table_data(table, identity)

                response_meta = { "response_op" : "ok", "identity" : util.Identity.get_identity() }
                return json.dumps(response_meta)

        ## Query op

        elif request.args['op'][0] == "query":

            # Check 'format' arg to determine output type.
            JSON = 1
            TEXT = 2
            HTML = 3  ## TODO
            format = JSON 
            try:
                if request.args['format'][0] == 'text':
                    format = TEXT
                elif request.args['format'][0] == 'html':
                    format = HTML
            except:
                pass

            # It is an error if no sql specified.
            if 'sql' not in request.args:
                logging.info("HttpReceiver: GET Query received with no sql.")
                if format == TEXT:
                    request.setHeader("Content-Type", "text/plain")
                    return "Must specify 'sql' param."
                else:
                    response_meta = { "response_op" : "error", "error_message" : "Must specify 'sql' param", "identity" : util.Identity.get_identity() }
                    return json.dumps(response_meta)


            sql = request.args['sql'][0]
            try:
                if format == TEXT:
                    self.process_sql_statement_for_text(sql, request)
                else:
                    self.process_sql_statement(sql, request)
                return NOT_DONE_YET

            except Exception as e:
                logging.error('HttpReceiver: Error processing sql: %s: %s', str(e), sql)
                if format == TEXT:
                    request.setHeader("Content-Type", "text/plain")
                    return "Error processing sql: %s" % str(e)
                else:
                    response_meta = { "response_op" : "error", "error_message" : "Error processing sql: %s" % str(e), "identity" : util.Identity.get_identity() }
                    return json.dumps(response_meta)

        logging.error("HttpReceiver: Unknown op: %s", request.args['op'][0])
        response_meta = { "response_op" : "error", "identity" : util.Identity.get_identity(), "error_message" : "Unrecognized operation" }
        return json.dumps(response_meta)

    def render_POST(self, request):

        request.setHeader("Content-Type", "application/json")

        #pprint(request.__dict__)
        obj = dict()

        if 'op' not in request.args:
            logging.error("HttpReceiver: No op specified for POST request: %s", ','.join(request.args))
            response_meta = { "response_op" : "error", "identity" : util.Identity.get_identity(), "error_message" : "No op specified" }
            return json.dumps(response_meta)

        ## Add CSV table data op

        if request.args['op'][0] == "add_csv_table_data":
            logging.info("HttpReceiver: Add table data (CSV).")

            csv = csv_table_reader.CsvTableReader()

            (table, error) = csv.read_table(request.content, None,
                                            skip_linenos={4},
                                            options_lineno=0,
                                            types_lineno=3,
                                            tablename_lineno=1,
                                            colnames_lineno=2)

            if table == None:
                logging.info("HttpReceiver: Failure parsing csv input: {}".format(error))
                response_meta = { "response_op" : "error", "identity" : util.Identity.get_identity(), "error_message" : str(error) }
                return json.dumps(response_meta)

            response_meta = { "response_op" : "ok", "identity" : util.Identity.get_identity() }

            try:
                if table.get_property("static"):
                    self.data_manager.sql_backend_writer_static.async_add_table_data(table, table.get_property("identity"))
                else:
                    self.data_manager.sql_backend_writer.async_add_table_data(table, table.get_property("identity"))
            except Exception as e:
                response_meta = { "response_op" : "error", "identity" : util.Identity.get_identity(), "error_message" : str(e) }
            
            return json.dumps(response_meta)

        ### The rest of the ops expect a JSON post body.

        try:
            # request.content might be a temp file if the content length is over some threshold.
            # fortunately json.load() will take a file ptr or a StringIO obj.
            obj = json.load(request.content)
            #print "Received POST:"
            #print obj

        except Exception as e:
            logging.info("HttpReceiver: ERROR failed to get/parse content of POST: %s", str(e))
            response_meta = { "response_op" : "error", "identity" : util.Identity.get_identity(), "error_message" : "Could not parse POST body: %s" % str(e) }
            return json.dumps(response_meta)

        ## Table list op

        if request.args['op'][0] == "get_table_list":
            #logging.info("HttpReceiver: Got request for table list.")
            response_meta = { "response_op" : "tables_list", "identity" : util.Identity.get_identity() }
            response_data = self.data_manager.get_table_list()
            return json.dumps(response_meta) + json.dumps(response_data)

        ## Add table data op

        if request.args['op'][0] == "add_table_data":

            #logging.info("HttpReceiver: Add table data (JSON).")
            table = qasino_table.QasinoTable()
            err = table.from_obj(obj)
            if err is not None:
                errmsg = "Invalid input format: " + str(err)
                logging.info("HttpReceiver: " + errmsg)
                response_meta = { "response_op" : "error", "identity" : util.Identity.get_identity(), "error_message" : errmsg }
            else:
                response_meta = { "response_op" : "ok", "identity" : util.Identity.get_identity() }
                try:
                    if table.get_property("static"):
                        self.data_manager.sql_backend_writer_static.async_add_table_data(table, table.get_property("identity"))
                    else:
                        self.data_manager.sql_backend_writer.async_add_table_data(table, table.get_property("identity"))
                except Exception as e:
                    response_meta = { "response_op" : "error", "identity" : util.Identity.get_identity(), "error_message" : str(e) }
            
            return json.dumps(response_meta)

        ## Query op

        if request.args['op'][0] == "query":

            if 'sql' not in obj:
                response_meta = { "response_op" : "error", "error_message" : "Must specify 'sql' param", "identity" : util.Identity.get_identity() }
                logging.info("HttpReceiver: Query received with no sql.")
                return json.dumps(response_meta)

            try:
                self.process_sql_statement(obj["sql"], request)
                return NOT_DONE_YET

            except Exception as e:
                logging.error('HttpReceiver: Error processing sql: %s: %s', str(e), obj["sql"])
                response_meta = { "response_op" : "error", "error_message" : str(e), "identity" : util.Identity.get_identity() }
                return json.dumps(response_meta)


        response_meta = { "response_op" : "error", "identity" : util.Identity.get_identity(), "error_message" : "Unrecognized operation" }
        return json.dumps(response_meta)


    def process_sql_statement(self, sql_statement, request):

        query_id = self.data_manager.get_query_id()

        query_start = time.time()

        logging.info("HttpReceiver: (%d) SQL received: %s", query_id, sql_statement.rstrip())

        if not sqlite3.complete_statement(sql_statement):

            # Try adding a semicolon at the end.
            sql_statement = sql_statement + ";"

            if not sqlite3.complete_statement(sql_statement):
                
                response_meta = { "response_op" : "error", 
                                  "error_message" : "Incomplete sql statement",
                                  "identity" : util.Identity.get_identity() }
                request.write(json.dumps(response_meta))
                request.finish()
                return

            # else it is now a complete statement

        d = self.data_manager.async_validate_and_route_query(sql_statement, query_id)

        d.addCallback(self.sql_complete_callback, query_id, query_start, request)

    def sql_complete_callback(self, result, query_id, query_start, request):
        """
        Called when a sql statement completes.
        """

        # To start just our identity.
        response_meta = { "identity" : util.Identity.get_identity() }

        retval = result["retval"]
        error_message = ''
        if "error_message" in result:
                error_message = str(result["error_message"])

        # Success?

        if retval == 0 and "data" in result:

            response_meta["response_op"] = "result_table"
            response_meta["table"] = result["data"]

            if "max_widths" in result:
                response_meta["max_widths"] = result["max_widths"]

        # Or error?

        if retval != 0:
            logging.info("HttpReceiver: (%d) SQL error: %s", query_id, error_message)

            response_meta["response_op"] = "error"
            response_meta["error_message"] = error_message

        else:
            logging.info("HttpReceiver: (%d) SQL completed (%.02f seconds)", query_id, time.time() - query_start)

        # Send the response!

        request.write(json.dumps(response_meta))
        request.finish()


    def process_sql_statement_for_text(self, sql_statement, request):

        query_id = self.data_manager.get_query_id()

        query_start = time.time()

        logging.info("HttpReceiver: (%d) SQL received: %s", query_id, sql_statement.rstrip())

        if not sqlite3.complete_statement(sql_statement):

            # Try adding a semicolon at the end.
            sql_statement = sql_statement + ";"

            if not sqlite3.complete_statement(sql_statement):
                
                request.write("Incomplete sql statement")
                request.finish()
                return

            # else it is now a complete statement

        d = self.data_manager.async_validate_and_route_query(sql_statement, query_id)

        d.addCallback(self.sql_complete_callback_for_text, query_id, query_start, request)

    def sql_complete_callback_for_text(self, result, query_id, query_start, request):
        """
        Called when a sql statement completes.
        """

        retval        = result["retval"]
        data          = result["data"]          if "data"          in result else None
        error_message = result["error_message"] if "error_message" in result else ''
        max_widths    = result["max_widths"]    if "max_widths"    in result else None

        if retval == 0 and data != None:

            # I'm sure there is a better way but I'm just not thinking straight right now..
            class Outputter(object):
                @staticmethod
                def sendLine(text):
                    request.write(text + "\n")

            outputter = Outputter()

            util.pretty_print_table(outputter, data, max_widths=max_widths)

        if retval != 0:
            logging.info("HttpReceiver: (%d) SQL error: %s", query_id, error_message)
            request.write("SQL error: %s" % error_message)
        else:
            logging.info("HttpReceiver: (%d) SQL completed (%.02f seconds)", query_id, time.time() - query_start)

        request.finish()




# Copyright (C) 2014 MediaMath, Inc. <http://www.mediamath.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from txzmq import ZmqFactory, ZmqEndpoint, ZmqEndpointType, ZmqPubConnection

import logging

import json

from util import Identity

class ZmqPublisher(ZmqPubConnection):

    def __init__(self, zmq_factory, port, data_manager=None):

        self.data_manager = data_manager

        endpoint = ZmqEndpoint(ZmqEndpointType.bind, "tcp://*:%d" % port)

        ZmqPubConnection.__init__(self, zmq_factory, endpoint)


    def send_generation_signal(self, generation_number, generation_duration_s):
        msg = { "op" : "generation_signal", "identity" : Identity.get_identity(), "generation_number" : generation_number }
        if generation_duration_s:
            msg["generation_duration_s"] = generation_duration_s

        self.publish(json.dumps(msg), "GENSIG")




# Copyright (C) 2014 MediaMath, Inc. <http://www.mediamath.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import random
import re

import util

class QasinoTable(object):
    """
    A simple container for a qasino table.
    """

    def __init__(self, tablename=None):
        self.tablename = tablename
        self.rows = []
        self.column_names = []
        self.column_types = []
        self.properties = {}

    def zip_columns(self):
        return zip(self.column_names, self.column_types)

    def set_tablename(self, tablename):
        self.tablename = tablename

    def get_tablename(self):
        return self.tablename

    def get_column_names(self):
        return self.column_names

    def init_retry(self, nr_retries=5):
        if self.__dict__.has_key('retry_count'):
            return
        self.retry_count = nr_retries

    def test_retry(self):
        try:
            self.retry_count -= 1
            if self.retry_count <= 0:
                return True
        except:
            pass

        return False

    def get_obj(self, **extra_settings):
        obj = {}
        for name, value in self.properties.iteritems():
            obj[name] = value

        for name, value in extra_settings.iteritems():
            obj[name] = value

        obj["table"] = { "tablename" : self.tablename,
                         "column_names" : self.column_names,
                         "column_types" : self.column_types,
                         "rows" : self.rows
                        }
        return obj

    def get_json(self, **extra_settings):
        return json.dumps(self.get_obj(**extra_settings))

    def get_rows(self):
        return self.rows

    def get_row(self, row_index):
        try:
            return self.rows[row_index]
        except:
            pass
        return []

    def validate(self):
        valid_column_name_regex = re.compile(r'^[a-zA-Z]+[a-zA-Z0-9_]*$')

        # Verify tablename is valid.
        if valid_column_name_regex.search(self.tablename) is None:
            raise Exception("Invalid tablename name '{}'".format(tablename))

        # Verify column names are valid.

        for column_name in self.column_names:
            if valid_column_name_regex.search(column_name) is None:
                raise Exception("Invalid column name '{}'".format(column_name))

        # Verify types are valid.
        valid_types = { x: 1 for x in [ 'text', 'varchar', 'real', 'integer', 'int' ] }

        for column_type in self.column_types:
            if column_type.lower() not in valid_types:
                raise Exception("Invalid column type '{}'".format(column_type))

        # We could/should check rows here probably...
        

    def from_obj(self, obj):
        try:
            self.tablename = obj['table']['tablename']
            self.column_names = obj['table']['column_names']
            self.column_types = obj['table']['column_types']
            self.rows = obj['table']['rows']

            for key, value in obj.iteritems():
                if key != "table" and key != "op":
                    self.properties[key] = value

            # Do some validation
            self.validate()

        except Exception as e:
            return e

        return None

    def printit(self):
        print self.get_obj()

    def get_nr_rows(self):
        return len(self.rows)

    def set_property(self, property_name, property_value):
        self.properties[property_name] = property_value

    def get_property(self, property_name):
        if property_name in self.properties:
            return self.properties[property_name]
        else:
            return None

    def set_column_names(self, column_names):
        self.column_names = column_names

    def set_column_types(self, column_types):
        self.column_types = column_types

    def add_column(self, colname, coltype):
        self.column_names.append(colname)
        self.column_types.append(coltype)

    def add_row(self, row):
        if len(row) != len(self.column_names) or len(row) != len(self.column_types):
            return -1
        self.rows.append(row)
        return 1

    def get_nr_column_names(self):
        return len(self.column_names)

    def get_nr_column_type(self):
        return len(self.column_types)



def get_a_random_table():

    type_array = [ "TEXT", "INTEGER", "REAL" ]

    nr_columns = random.randint(1, 20)

    column_types = [ type_array[ random.randint(0, len(type_array) - 1) ] for _ in range(nr_columns) ]

    table = QasinoTable(util.random_string(4, 20))
    table.set_column_names( [ util.random_string(1, 40) for _ in range(nr_columns) ] )
    table.set_column_types( column_types )

    rows = []
    for row_index in range(random.randint(1, 300)):
        row = []
        for column_index in range(nr_columns):
            if column_types[column_index] == "TEXT":
                row.append(util.random_string(1, 50))
            elif column_types[column_index] == "REAL":
                row.append(random.randint(0, 3483839392) / 100.0)
            else:
                row.append(random.randint(0, 3483839392))
        table.add_row(row)

    return table



# Copyright (C) 2014 MediaMath, Inc. <http://www.mediamath.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sqlite_backend as sql_backend

import table_merger
import util
import qasino_table

import logging
import time
import re
import yaml
import sys
import thread

from twisted.internet import threads
from twisted.internet import task

class DataManager(object):

    def __init__(self, use_dbfile, db_dir=None, signal_channel=None, archive_db_dir=None, 
                 generation_duration_s=30):

        self.saved_tables = {}
        self.query_id = 0
        self.views = {}
        self.thread_id = thread.get_ident()
        self.stats = {}

        self.generation_duration_s = generation_duration_s
        self.signal_channel = signal_channel
        self.archive_db_dir = archive_db_dir
        self.static_db_filepath = db_dir + '/qasino_table_store_static.db'

        # Start with zero because we'll call rotate_dbs instantly below.
        self.db_generation_number = 0

        # use_dbfile can be:
        #   'memory- -> use in memory db
        #   /%d/ -> use the provided template filename
        #   ! /%d/ -> use the filename (same db every generation)

        self.one_db = False
        self.db_name = use_dbfile

        if use_dbfile == None:
            self.db_name = "qasino_table_store_%d.db"

        elif use_dbfile == "memory":
            self.db_name = ":memory:"
            self.one_db = True

        elif use_dbfile.find('%d') == -1:
            self.one_db = True

        # Initialize some things

        self.table_merger = table_merger.TableMerger(self)

        # Add db_dir path

        if db_dir != None and self.db_name != ":memory:":
            self.db_name = db_dir + '/' + self.db_name

        # Open the writer backend db.

        db_file_name = self.db_name

        if not self.one_db:
            db_file_name = self.db_name % self.db_generation_number

        self.sql_backend_reader = None
        self.sql_backend_writer = sql_backend.SqlConnections(db_file_name, 
                                                             self, 
                                                             self.archive_db_dir, 
                                                             self.thread_id, 
                                                             self.static_db_filepath)

        self.sql_backend_writer_static = sql_backend.SqlConnections(self.static_db_filepath, 
                                                                    self, 
                                                                    self.archive_db_dir, 
                                                                    self.thread_id, 
                                                                    None)
                                                                    

        # Make the data manager db rotation run at fixed intervals.
        # This will also immediately make the call which will make the
        # writer we just opened the reader and to open a new writer.

        self.rotate_task = task.LoopingCall(self.async_rotate_dbs)
        self.rotate_task.start(self.generation_duration_s)


    def read_views(self, filename):

        # Reset views
        self.views = {}

        try:
            fh = open(filename, "r")
        except Exception as e:
            logging.info("Failed to open views file '%s': %s", filename, e)
            return

        try:
            view_conf_obj = yaml.load(fh)
        except Exception as e:
            logging.info("Failed to parse view conf yaml file '%s': %s", filename, e)
            return

        for view in view_conf_obj:

            try:
                viewname = view["viewname"]
                view = view["view"]
                self.views[viewname] = { 'view' : view, 'loaded' : False, 'error' : '' }
            except Exception as e:
                logging.info("Failure getting view '%s': %s", view["viewname"] if "viewname" in view else 'unknown', e)

    def get_query_id(self):
        self.query_id += 1
        return self.query_id

    def shutdown(self):
        self.rotate_task = None
        self.sql_backend_reader = None
        self.sql_backend_writer = None


    def async_validate_and_route_query(self, sql, query_id, use_write_db=False):
        if use_write_db:
            return self.sql_backend_writer.run_interaction(sql_backend.SqlConnections.WRITER_INTERACTION, 
                                                           self.validate_and_route_query, sql, query_id, self.sql_backend_writer)
        else:
            return self.sql_backend_reader.run_interaction(sql_backend.SqlConnections.READER_INTERACTION,
                                                           self.validate_and_route_query, sql, query_id, self.sql_backend_reader)


    def validate_and_route_query(self, txn, sql, query_id, sql_backend):

        m = re.search(r"^\s*select\s+", sql, flags=re.IGNORECASE)
        if m == None:

            # Process a non-select statement.

            return self.process_non_select(txn, sql, query_id, sql_backend)

        # Process a select statement.

        return sql_backend.do_select(txn, sql)


    def process_non_select(self, txn, sql, query_id, sql_backend):
        """
        Called for non-select statements like show tables and desc.
        """

        # DESC?

        m = re.search(r"^\s*desc\s+(\S+)\s*;$", sql, flags=re.IGNORECASE)

        if m != None:
            (retval, error_message, table) = sql_backend.do_desc(txn, m.group(1))

            result = { "retval" : retval }
            if error_message:
                result["error_message"] = error_message
            if table:
                result["data"] = table

            return result

        # DESC VIEW?

        m = re.search(r"^\s*desc\s+view\s+(\S+)\s*;$", sql, flags=re.IGNORECASE)
        if m != None:
            return sql_backend.do_select(txn, "SELECT view FROM qasino_server_views WHERE viewname = '%s';" % m.group(1))
            
        # SHOW tables?

        m = re.search(r"^\s*show\s+tables\s*;$", sql, flags=re.IGNORECASE)
        if m != None:
            return sql_backend.do_select(txn, "SELECT *, strftime('%Y-%m-%d %H:%M:%f UTC', last_update_epoch, 'unixepoch') last_update_datetime FROM qasino_server_tables ORDER BY tablename;")

        # SHOW tables with LIKE?

        m = re.search(r"^\s*show\s+tables\s+like\s+('\S+')\s*;$", sql, flags=re.IGNORECASE)
        if m != None:
            return sql_backend.do_select(txn, "SELECT *, strftime('%Y-%m-%d %H:%M:%f UTC', last_update_epoch, 'unixepoch') last_update_datetime FROM qasino_server_tables WHERE tablename LIKE {} ORDER BY tablename;".format(m.group(1)) )

        # SHOW connections?

        m = re.search(r"^\s*show\s+connections\s*;$", sql, flags=re.IGNORECASE)
        if m != None:
            return sql_backend.do_select(txn, "SELECT *, strftime('%Y-%m-%d %H:%M:%f UTC', last_update_epoch, 'unixepoch') last_update_datetime FROM qasino_server_connections ORDER BY identity;")

        # SHOW info?

        m = re.search(r"^\s*show\s+info\s*;$", sql, flags=re.IGNORECASE)
        if m != None:
            return sql_backend.do_select(txn, "SELECT *, strftime('%Y-%m-%d %H:%M:%f UTC', generation_start_epoch, 'unixepoch') generation_start_datetime FROM qasino_server_info;")

        # SHOW views?

        m = re.search(r"^\s*show\s+views\s*;$", sql, flags=re.IGNORECASE)
        if m != None:
            return sql_backend.do_select(txn, "SELECT viewname, loaded, errormsg FROM qasino_server_views ORDER BY viewname;")

        # Exit?

        m = re.search(r"^\s*(quit|logout|exit)\s*;$", sql, flags=re.IGNORECASE)
        if m != None:
            return { "retval" : 0, "error_message" : "Bye!" }


        return { "retval" : 1, "error_message" : "ERROR: Unrecognized statement: %s" % sql }


    def get_table_list(self):

        return self.sql_backend_reader.tables

    def insert_tables_table(self, txn, sql_backend_writer, sql_backend_writer_static):

        table = qasino_table.QasinoTable("qasino_server_tables")
        table.add_column("tablename",         "varchar")
        table.add_column("nr_rows",           "int")
        table.add_column("nr_updates",        "int")
        table.add_column("last_update_epoch", "int")
        table.add_column("static",            "int")

        sql_backend_writer.add_tables_table_rows(table)

        sql_backend_writer_static.add_tables_table_rows(table)

        # the chicken or the egg - how do we add ourselves?

        table.add_row( [ "qasino_server_tables",
                         table.get_nr_rows() + 1,
                         1,
                         time.time(), 
                         0 ] )

        return sql_backend_writer.add_table_data(txn, table, util.Identity.get_identity())
        

    # This hack insures all the internal tables are inserted
    # using the same sql_backend_writer and makes sure that the
    # "tables" table is called last (after all the other internal
    # tables are added).

    def insert_internal_tables(self, txn, sql_backend_writer, sql_backend_reader, db_generation_number, time, generation_duration_s, views):

        sql_backend_writer.insert_info_table(txn, db_generation_number, time, generation_duration_s)

        sql_backend_writer.insert_connections_table(txn)

        if sql_backend_reader != None:
            sql_backend_writer.insert_sql_stats_table(txn, sql_backend_reader)

        sql_backend_writer.insert_update_stats_table(txn)

        # this should be second last so views can be created of any tables above.
        # this means though that you can not create views of any tables below.
        sql_backend_writer.add_views(txn, views)

        sql_backend_writer.insert_views_table(txn, views)

        # this should be last to include all the above tables

        self.insert_tables_table(txn, sql_backend_writer, self.sql_backend_writer_static)

    def async_rotate_dbs(self):
        """
        Kick off the rotate in a sqlconnection context because we have
        some internal tables and views to add before we rotate dbs.
        """
        
        self.sql_backend_writer.run_interaction(sql_backend.SqlConnections.WRITER_INTERACTION, self.rotate_dbs)

    def rotate_dbs(self, txn):
        """ 
        Make the db being written to be the reader db.
        Open a new writer db for all new updates.
        """

        logging.info("**** DataManager: Starting generation %d", self.db_generation_number)

        # Before making the write db the read db, 
        # add various internal info tables and views.


        self.insert_internal_tables(txn, 
                                    self.sql_backend_writer, 
                                    self.sql_backend_reader,
                                    self.db_generation_number, 
                                    time.time(), 
                                    self.generation_duration_s, 
                                    self.views)

        # Increment the generation number.

        self.db_generation_number = int(time.time())

        # Set the writer to a new db

        save_sql_backend_writer = self.sql_backend_writer

        # If specified put the generation number in the db name.

        db_file_name = self.db_name

        if not self.one_db:
                db_file_name = self.db_name % self.db_generation_number

        self.sql_backend_writer = sql_backend.SqlConnections(db_file_name, 
                                                             self, 
                                                             self.archive_db_dir, 
                                                             self.thread_id, 
                                                             self.static_db_filepath)

        # Set the reader to what was the writer
        # Note the reader will (should) be deconstructed here.

        self.sql_backend_reader = save_sql_backend_writer

        # Load saved tables.

        self.async_add_saved_tables()

        # Lastly blast out the generation number.

        if self.signal_channel != None:
            self.signal_channel.send_generation_signal(self.db_generation_number, self.generation_duration_s)


    def check_save_table(self, table, identity):

        tablename = table.get_tablename()

        key = tablename + identity

        if table.get_property('persist'):
            self.saved_tables[key] = { "table" : table, "tablename" : tablename, "identity" : identity }
        else:
            # Be sure to remove a table that is no longer persisting.
            if key in self.saved_tables:
                del self.saved_tables[key]


    def async_add_saved_tables(self):

        for key, table_data in self.saved_tables.iteritems():

            logging.info("DataManager: Adding saved table '%s' from '%s'", table_data["tablename"], table_data["identity"])

            self.sql_backend_writer.async_add_table_data(table_data["table"], table_data["identity"])



#!/usr/bin/python

# Copyright (C) 2014 MediaMath, Inc. <http://www.mediamath.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
from optparse import OptionParser

for path in [
    os.path.join('opt', 'qasino', 'lib'),
    os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'lib'))
]:
    if os.path.exists(os.path.join(path, '__init__.py')):
        sys.path.insert(0, path)
        break

import constants
from util import Identity
import qasino_table

def read_json_table(filehandle, options):
    """
    Read json table from the input file handle.
    """

    json_str = filehandle.read()

    try:
        import json

        json_obj = json.loads(json_str)
    except Exception as e:
        print "Failed to parse input json: {}".format(e)
        return

    table = qasino_table.QasinoTable()
    table.from_obj( { "table" : json_obj } )

    return table
    

def read_csv_table(filehandle, options):
    """
    Read csv table from the input file handle.
    """

    from csv_table_reader import CsvTableReader

    nr_errors = 0

    # Create a csv table reader object.

    csv_table_reader = CsvTableReader()

    # The csv files we'll be reading in will have this format:
    # ------------------------
    # version/options
    # tablename
    # column names (csv)
    # column types (csv)
    # column descriptions (csv)
    # data (csv)
    # ...
    # ------------------------

    # Ignore the 2nd and 5th lines.  Names in 3rd, types in 4th.  The "version" in the first line is now the options list.
    (table, error) = csv_table_reader.read_table(filehandle, None,
                                                 skip_linenos={4},
                                                 options_lineno=0,
                                                 types_lineno=3,
                                                 tablename_lineno=1,
                                                 colnames_lineno=2)

    if table == None:
        print "Failure reading csv file: {}".format(error)
        return None

    return table



def main():

    #logging.basicConfig(format="%(asctime)s %(message)s", datefmt="%Y-%m-%d %H:%M:%S",
    #                    level=logging.INFO)

    parser = OptionParser()

    parser.add_option("-I", "--identity", dest="identity",
                      help="Use IDENTITY as identity", metavar="IDENTITY")

    parser.add_option("-H", "--hostname", dest="hostname", default='localhost',
                      help="Send table to HOSTNAME qasino server", metavar="HOSTNAME")

    parser.add_option("-p", "--port", dest="port", default=-1,
                      help="Use PORT to connect to", metavar="PORT")

    parser.add_option("-u", "--username", dest="username", 
                      help="HTTPS auth username")

    parser.add_option("-w", "--password", dest="password", 
                      help="HTTPS auth password")

    parser.add_option("-P", "--persist", dest="persist", default=False, action="store_true",
                      help="Set the persistent option on the table")

    parser.add_option("-S", "--static", dest="static", default=False, action="store_true",
                      help="Set the static option on the table")

    parser.add_option("-t", "--use-http", dest="use_http", default=False, action="store_true",
                      help="Use HTTP protocol to publish table.")

    parser.add_option("-s", "--use-https", dest="use_https", default=False, action="store_true",
                      help="Use HTTP over SSL/TLS protocol to publish table.")

    parser.add_option("-z", "--use-zmq", dest="use_zmq", default=False, action="store_true",
                      help="Use ZeroMQ protocol to publish table.")

    parser.add_option("-k", "--skip-ssl-verify", dest="skip_ssl_verify", default=False, action="store_true",
                      help="Don't verify SSL certificates.")

    parser.add_option("-d", "--data-format", dest="data_format", default='json',
                      help="Read data in as either 'json' or 'csv'.")

    parser.add_option("-f", "--filename", dest="filename", 
                      help="Read data from FILENAME.")


    (options, args) = parser.parse_args()

    print "Qasino publish starting"

    if options.identity != None:
        Identity.set_identity(options.identity)

    print "Identity is {}".format(Identity.get_identity())

    if options.hostname == None:
        print "Please specify a hostname to connect to."
        exit(1)

    # invalid combos
    if (options.use_zmq and options.use_https) or (options.use_http and options.use_https) or (options.use_http and options.use_zmq):
        print "Pick one of --use-https, --use-https or --use-zmq"
        exit(1);

    # default to https
    if not options.use_http and not options.use_https and not options.use_zmq:
        options.use_https = True

    # set the default port
    if options.port == -1:
        if options.use_http:
            options.port = constants.HTTP_PORT
        elif options.use_https:
            options.port = constants.HTTPS_PORT
        elif options.use_zmq:
            options.port = constants.ZMQ_RPC_PORT

    # Read from stdin by default or filename if specified.
    filename = options.filename if options.filename else "/dev/stdin"

    try:
        fh = open(filename)
    except Exception as e:
        print "Failed to open '{}': {}".format(filename, str(e))
        exit(1)

    # Are we reading a csv table or a json table?
    if options.data_format == 'json':
        table = read_json_table(fh, options)

    elif options.data_format == 'csv':
        table = read_csv_table(fh, options)

    fh.close()

    if table == None:
        print "Failed to get a valid table!"
        exit(1)

    if options.persist:
        table.set_property("persist", 1)

    if options.static:
        table.set_property("static", 1)

    properties = []
    if table.get_property('static'): properties.append(' static')
    if table.get_property('update'): properties.append(' update')
    if table.get_property('persist'): properties.append(' persist')

    print "Sending{} table '{}' to '{}:{}' ({} rows).".format(''.join(properties), table.get_tablename(), options.hostname, options.port, table.get_nr_rows())

    if options.use_zmq:

        from txzmq import ZmqFactory
        import zmq_requestor

        zmq_factory = ZmqFactory()

        # Create a json request object.

        zmq_requestor = zmq_requestor.ZmqRequestor(options.hostname, options.port, zmq_factory)

        zmq_requestor.send_table(table)


    elif options.use_http or options.use_https:

        if options.use_https:
            url_proto = 'https'
            if options.port == constants.HTTP_PORT:
                options.port = constants.HTTPS_PORT
        else:
            url_proto = 'http'

        import requests

        conn = requests.Session()

        request_options = { 'headers' : {'Content-Type': 'application/json'} }

        if options.skip_ssl_verify:
            request_options['verify'] = False
        if options.username and options.password:
            request_options['auth'] = (options.username, options.password)

        URL = '{}://{}:{}/request?op=add_table_data'.format(url_proto, options.hostname, options.port)

        jsondata = table.get_json(op="add_table_data", identity=Identity.get_identity())

        #print jsondata
    
        request_options['data'] = jsondata

        response = conn.post(URL, **request_options)
        try:
            response.raise_for_status()
        except Exception as e:
            print "ERROR: Request failed: {}".format(e)

    else:

        print "ERROR: Unknown or No transport specified."
        exit(1)

    print "Qasino publish exiting"
    exit(0)

    # END main
    
if __name__ == '__main__':
    main()



from setuptools import setup
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()
    
setup(
    name='qasino_client',
    version='1.6.2',
    
    description='A command line client for the qasino project.',
    long_description=long_description,
    
    url='https://github.com/MediaMath/qasino',
    license='Apache 2.0',
    
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: System :: Monitoring',
        'Programming Language :: Python :: 2.7',
    ],
    
    keywords='qasino stats monitoring mediamath',
    
    package_dir = {'qasino_client': ''},
    packages=['qasino_client', 'qasino_client.lib'],
    py_modules=['qasino_client.bin.qasino_sqlclient',
                'qasino_client.bin.qasino_csvpublisher'],
    scripts=['bin/qasinosql',
             'bin/qasinocsvpub'],
    install_requires=['requests', 'txzmq'],
)



# Copyright (C) 2014 MediaMath, Inc. <http://www.mediamath.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests
from util import Identity

class HttpRequestor():

    def __init__(self, hostname, port, username=None, password=None, skip_ssl_verify=False, url_proto='https'):
        self.hostname = hostname
        self.port = port
        self.username = username
        self.password = password
        self.skip_ssl_verify = skip_ssl_verify
        self.url_proto = url_proto
        self.conn = requests.Session()

    def send_table(self, table):

        request_options = { 'headers' : {'Content-Type': 'application/json'} }

        if self.skip_ssl_verify:
            request_options['verify'] = False
        if self.username and self.password:
            request_options['auth'] = (self.username, self.password)

        url = '{}://{}:{}/request?op=add_table_data'.format(self.url_proto, self.hostname, self.port)

        jsondata = table.get_json(op="add_table_data", identity=Identity.get_identity())

        #print jsondata
    
        request_options['data'] = jsondata

        try:
            response = self.conn.post(url, **request_options)
            response.raise_for_status()
        except Exception as e:
            return "ERROR: HttpRequestor: Request failed: url={}: {}".format(url, e)

        return None



# Copyright (C) 2014 MediaMath, Inc. <http://www.mediamath.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from txzmq import ZmqFactory, ZmqEndpoint, ZmqEndpointType, ZmqSubConnection

import logging

import json

from util import Identity

class ZmqSubscriber(ZmqSubConnection):

    def __init__(self, remote_host, port, zmq_factory, data_manager=None):

        self.data_manager = data_manager
        self.remote_host = remote_host
        self.generation_signal_listeners = []

        endpoint = ZmqEndpoint(ZmqEndpointType.connect, "tcp://%s:%d" % (remote_host, port))

        ZmqSubConnection.__init__(self, zmq_factory, endpoint)

    def shutdown(self):
        self.unsubscribe("GENSIG")
        ZmqSubConnection.shutdown(self)

    def subscribe_generation_signal(self, callback, *args):

        self.generation_signal_listeners.append( (callback, args) )
        self.subscribe("GENSIG")

    def gotMessage(self, message, tag):
        
        if tag == "GENSIG":
            logging.info("Received generation signal!")
            for callback, args in self.generation_signal_listeners:
                callback(*args)



# Copyright (C) 2014 MediaMath, Inc. <http://www.mediamath.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

class TableMerger(object):

    def __init__(self, data_manager):
        self.data_manager = data_manager


    def merge_table(self, txn, table_to_add, existing_schema, sql_backend):

        # Determine if we need to merge this table.

        tablename = table_to_add.get_tablename()

        # Make some sets

        table_to_add_columns = set( table_to_add.get_column_names() )
        existing_columns = set( [ row[0] for row in existing_schema ] )

        # Find discrepancies between table and existing schema.

        # First, are there new columns in the table to add?

        columns_to_add = table_to_add_columns - existing_columns

        if len(columns_to_add) > 0:

            # The existing table we are adding to is missing columns.
            # Try to add the missing columns to the table.

            column_type_lookup = { column_name : column_type 
                                   for column_name, column_type in table_to_add.zip_columns() }

            for column_name in columns_to_add:
                column_type = column_type_lookup[column_name]
                sql = "ALTER TABLE '%s' ADD COLUMN '%s' %s DEFAULT NULL;" % (tablename, column_name, column_type)

                logging.info("TableMerger: Altering table %s to add column %s %s", tablename, column_name, column_type)
                sql_backend.do_sql(txn, sql)


        # Else we may have columns that are missing from the table to
        # add (ie in the existing schema but not in the new table to
        # add).  We do nothing though because the insert will just not
        # insert those columns and we assume default values for all
        # columns.

        # TODO type changes.

